from tupa.action import Action, Actions
from tupa.config import Config
from ucca import layer1


class Oracle(object):
    """
    Oracle to produce gold transition parses given UCCA passages
    To be used for creating training data for a transition-based UCCA parser
    :param passage gold passage to get the correct edges from
    """
    def __init__(self, passage):
        l1 = passage.layer(layer1.LAYER_ID)
        self.nodes_remaining = {node.ID for node in l1.all
                                if node is not l1.heads[0] and
                                (Config().args.linkage or node.tag != layer1.NodeTags.Linkage) and
                                (Config().args.implicit or not node.attrib.get("implicit"))}
        self.edges_remaining = {edge for node in passage.nodes.values() for edge in node
                                if (Config().args.linkage or edge.tag not in (
                                    layer1.EdgeTags.LinkRelation, layer1.EdgeTags.LinkArgument)) and
                                (Config().args.implicit or not edge.child.attrib.get("implicit")) and
                                (Config().args.remote or not edge.attrib.get("remote"))}
        self.passage = passage
        self.edge_found = False
        self.log = None

    def get_actions(self, state):
        """
        Determine all zero-cost action according to current state
        Asserts that the returned action is valid before returning
        :param state: current State of the parser
        :return: list of Action items to perform
        """
        actions = []
        invalid = []
        for action in self.generate_actions(state):
            action.generate_id()
            try:
                state.assert_valid(action)
                actions.append(action)
            except AssertionError as e:
                invalid.append((action, e))
        assert actions, self.generate_log(invalid, state)
        return actions

    def generate_log(self, invalid, state):
        self.log = "\n".join(["Oracle found no valid action",
                              state.str("\n"), self.str("\n"),
                              "Actions returned by the oracle:"] +
                             ["  %s: %s" % (action, e) for (action, e) in invalid])
        return self.log

    def generate_actions(self, state):
        """
        Determine all zero-cost action according to current state
        :param state: current State of the parser
        :return: generator of Action items to perform
        """
        if not self.edges_remaining or not state.buffer and not state.stack:
            yield Actions.Finish
            if state.stack:
                yield Actions.Reduce
            return

        self.edge_found = False
        if state.stack:
            s0 = state.stack[-1]
            incoming = self.edges_remaining.intersection(s0.orig_node.incoming)
            outgoing = self.edges_remaining.intersection(s0.orig_node.outgoing)
            if not incoming and not outgoing:
                yield Actions.Reduce
                return
            else:
                # Check for actions to create new nodes
                for edge in incoming:
                    if edge.parent.ID in self.nodes_remaining and not edge.attrib.get("remote"):
                        yield self.create_node_action(edge, edge.parent, Actions.Node)

                for edge in outgoing:
                    if edge.child.attrib.get("implicit"):
                        yield self.create_node_action(edge, edge.child, Actions.Implicit)

                if len(state.stack) > 1:
                    s1 = state.stack[-2]
                    # Check for actions to create binary edges
                    for edge in incoming:
                        if edge.parent.ID == s1.node_id:
                            yield self.create_edge_action(edge, Action.RIGHT)

                    for edge in outgoing:
                        if edge.child.ID == s1.node_id:
                            yield self.create_edge_action(edge, Action.LEFT)

                    if not self.edge_found:
                        # Check if a swap is necessary, and how far (if compound swap is enabled)
                        related = dict([(edge.child.ID,  edge) for edge in outgoing] +
                                       [(edge.parent.ID, edge) for edge in incoming])
                        distance = None  # Swap distance (how many nodes in the stack to swap)
                        for i, s in enumerate(state.stack[-3::-1]):  # Skip top two, they are not related
                            edge = related.pop(s.node_id, None)
                            if edge is not None:
                                if not Config().args.swap:  # We have no chance to reach it, so stop trying
                                    self.remove(edge)
                                    continue
                                if distance is None and Config().args.compound_swap:  # Save the first one
                                    distance = i + 1
                                if not related:  # All related nodes are in the stack
                                    yield Actions.Swap(distance)
                                    return

        if state.buffer and not self.edge_found:
            yield Actions.Shift

    def create_edge_action(self, edge, direction):
        self.edge_found = True
        return Action.edge_action(direction, edge.attrib.get("remote"), edge.tag,
                                  orig_edge=edge, oracle=self)

    def create_node_action(self, edge, node, action):
        self.edge_found = True
        return action(edge.tag, orig_edge=edge, orig_node=node, oracle=self)

    def remove(self, edge, node=None):
        self.edges_remaining.discard(edge)
        if node is not None:
            self.nodes_remaining.discard(node.ID)

    def str(self, sep):
        return "nodes left: [%s]%sedges left: [%s]" % (" ".join(self.nodes_remaining), sep,
                                                       " ".join(map(str, self.edges_remaining)))

    def __str__(self):
        return str(" ")



import time

from classifiers.classifier import Classifier, ClassifierProperty


class Perceptron(Classifier):
    """
    Abstract multi-class averaged perceptron.
    """

    def __init__(self, *args, model=None, epoch=0):
        """
        Create a new untrained Perceptron or copy the weights from an existing one
        :param model: if given, copy the weights (from a trained model)
        """
        super(Perceptron, self).__init__(*args, model=model)
        if self.is_frozen:
            self.model = model
        self._update_index = 0  # Counter for calls to update()
        self.initial_learning_rate = self.learning_rate
        self.epoch = epoch
        self.update_learning_rate()

    def update(self, features, pred, true, importance=1):
        """
        Update classifier weights according to predicted and true labels
        :param features: extracted feature values, of size num_features
        :param pred: label predicted by the classifier (non-negative integer less than num_labels)
        :param true: true label (non-negative integer less than num_labels)
        :param importance: how much to scale the feature vector for the weight update
        """
        super(Perceptron, self).update(features, pred, true, importance)
        self._update_index += 1

    def finalize(self, average=True, finished_epoch=False):
        """
        Average all weights over all updates, as a form of regularization
        :param average: whether to really average the weights or just return them as they are now
        :param finished_epoch: whether to decay the learning rate
        :return new Perceptron object with the weights averaged
        """
        super(Perceptron, self).finalize()
        started = time.time()
        if finished_epoch:
            self.epoch += 1
        if average:
            print("Averaging weights... ", end="", flush=True)
        finalized = self._finalize_model(average)
        if average:
            print("Done (%.3fs)." % (time.time() - started))
        print("Labels: %d" % self.num_labels)
        print("Features: %d" % self.input_dim)
        return finalized

    def _finalize_model(self, average):
        raise NotImplementedError()

    def update_learning_rate(self):
        self.learning_rate = self.initial_learning_rate / (1.0 + self.epoch * self.learning_rate_decay)

    def resize(self):
        raise NotImplementedError()

    def save_model(self):
        """
        Save all parameters to file
        """
        d = {
            "_update_index": self._update_index,
            "initial_learning_rate": self.initial_learning_rate,
            "epoch": self.epoch,
        }
        d.update(self.save_extra())
        return d

    def save_extra(self):
        return {"model": self.model}

    def load_model(self, d):
        self._update_index = d["_update_index"]
        self.initial_learning_rate = d["initial_learning_rate"]
        self.epoch = d["epoch"]
        self.load_extra(d)

    def load_extra(self, d):
        self.model = d["model"]

    def write(self, filename, sep="\t"):
        print("Writing model to '%s'..." % filename)
        with open(filename, "w") as f:
            self.write_model(f, sep)

    def write_model(self, f, sep):
        raise NotImplementedError()

    def get_classifier_properties(self):
        return super(Perceptron, self).get_classifier_properties() + \
               (ClassifierProperty.update_only_on_error,)



from enum import Enum

from tupa.config import Config
from tupa.model_util import load_dict, save_dict


class ClassifierProperty(Enum):
    update_only_on_error = 1
    require_init_features = 2
    trainable_after_saving = 3


class Classifier(object):
    """
    Interface for classifier used by the parser.
    """

    def __init__(self, model_type, filename, labels, model=None):
        """
        :param labels: a list of labels that can be updated later to add a new label
        :param model: if given, copy the weights (from a trained model)
        """
        self.model = None
        self.model_type = model_type
        self.filename = filename
        self.labels = labels or []
        self._num_labels = len(self.labels)
        self.is_frozen = model is not None
        self.learning_rate = Config().args.learning_rate
        self.learning_rate_decay = Config().args.learning_rate_decay

    @property
    def num_labels(self):
        return len(self.labels)

    def score(self, features):
        if not self.is_frozen:
            self._update_num_labels()

    def init_features(self, features, train=False):
        pass

    def update(self, features, pred, true, importance=1):
        assert not self.is_frozen, "Cannot update a frozen model"
        self._update_num_labels()

    def _update_num_labels(self):
        """
        self.num_labels is updated automatically when a label is added to self.labels,
        but we need to update the weights whenever that happens
        """
        if self.num_labels > self._num_labels:
            self._num_labels = self.num_labels
            self.resize()

    def resize(self):
        raise NotImplementedError()

    def finalize(self, *args, **kwargs):
        assert not self.is_frozen, "Cannot freeze a frozen model"
        self._update_num_labels()

    def finished_step(self, train=False):
        """
        Called by the parser when a single step is finished
        """
        pass

    def finished_item(self, train=False):
        """
        Called by the parser when a whole item is finished
        """
        pass

    def save(self):
        """
        Save all parameters to file
        """
        d = {
            "type": self.model_type,
            "labels": self.labels,
            "is_frozen": self.is_frozen,
            "learning_rate": self.learning_rate,
            "learning_rate_decay": self.learning_rate_decay,
        }
        d.update(self.save_model())
        save_dict(self.filename, d)

    def save_model(self):
        return self.save_extra()

    def save_extra(self):
        return {}

    def load(self):
        """
        Load all parameters from file
        """
        d = load_dict(self.filename)
        model_type = d.get("type")
        assert model_type is None or model_type == self.model_type, \
            "Model type does not match: %s" % model_type
        self.labels = list(d["labels"])
        self.is_frozen = d["is_frozen"]
        self.learning_rate = d["learning_rate"]
        self.learning_rate_decay = d["learning_rate_decay"]
        self.load_model(d)

    def load_model(self, d):
        self.load_extra(d)

    def load_extra(self, d):
        pass

    def get_classifier_properties(self):
        return ()

    def __str__(self):
        return ("%d labels, " % self.num_labels) + (
                "%d features" % self.input_dim)



import argparse
import sys

import numpy as np

from ucca import convert, constructions


class Singleton(type):
    instance = None

    def __call__(cls, *args, **kwargs):
        if not cls.instance:
            cls.instance = super(Singleton, cls).__call__(*args, **kwargs)
        return cls.instance

    def reload(cls):
        cls.instance = None


class VAction(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        if values is None:
            values = "1"
        try:
            values = int(values)
        except ValueError:
            values = values.count("v") + 1
        setattr(args, self.dest, values)

SPARSE_PERCEPTRON = "sparse"
DENSE_PERCEPTRON = "dense"
MLP_NN = "mlp"
BILSTM_NN = "bilstm"
CLASSIFIERS = (SPARSE_PERCEPTRON, DENSE_PERCEPTRON, MLP_NN, BILSTM_NN)

# Multiple choice options: the first one is always the default
ACTIVATIONS = ("sigmoid", "tanh", "relu", "cube")
INITIALIZATIONS = ("glorot_uniform", "normal", "uniform", "const")
OPTIMIZERS = ("adam", "sgd", "momentum", "adagrad", "adadelta")


class Config(object, metaclass=Singleton):
    def __init__(self, *args):
        argparser = argparse.ArgumentParser(description="""Transition-based parser for UCCA.""",
                                            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        argparser.add_argument("passages", nargs="*", help="passage files/directories to test on/parse")
        argparser.add_argument("-m", "--model", help="model file to load/save (default: ucca_<model_type>")
        argparser.add_argument("-c", "--classifier", choices=CLASSIFIERS, default=SPARSE_PERCEPTRON, help="model type")
        argparser.add_argument("-B", "--beam", choices=(1,), default=1, help="beam size for beam search (1 for greedy)")
        argparser.add_argument("-e", "--evaluate", action="store_true", help="evaluate parsed passages")
        argparser.add_argument("-v", "--verbose", nargs="?", action=VAction, default=0, help="detailed parse output")
        constructions.add_argument(argparser)
        group = argparser.add_mutually_exclusive_group()
        group.add_argument("-s", "--sentences", action="store_true", help="separate passages to sentences")
        group.add_argument("-a", "--paragraphs", action="store_true", help="separate passages to paragraphs")
        group = argparser.add_argument_group(title="Training parameters")
        group.add_argument("-t", "--train", nargs="+", default=(), help="passage files/directories to train on")
        group.add_argument("-d", "--dev", nargs="+", default=(), help="passage files/directories to tune on")
        group.add_argument("-I", "--iterations", type=int, default=1, help="number of training iterations")
        group.add_argument("--folds", type=int, choices=(3, 5, 10), help="#folds for cross validation")
        group.add_argument("--seed", type=int, default=1, help="random number generator seed")
        group = argparser.add_argument_group(title="Output files")
        group.add_argument("-o", "--outdir", default=".", help="output directory for parsed files")
        group.add_argument("-p", "--prefix", default="", help="output filename prefix")
        group.add_argument("-W", "--no-write", action="store_true", help="do not write parsed passages to files")
        group.add_argument("-l", "--log", help="output log file (default: model filename + .log)")
        group.add_argument("--devscores", help="output CSV file for dev scores (default: model filename + .dev.csv)")
        group.add_argument("--testscores", help="output CSV file for test scores (default: model filename + .test.csv)")
        group = argparser.add_argument_group(title="Structural constraints")
        group.add_argument("--linkage", action="store_true", help="include linkage nodes and edges")
        group.add_argument("--implicit", action="store_true", help="include implicit nodes and edges")
        group.add_argument("--no-remote", action="store_false", dest="remote", help="ignore remote edges")
        group.add_argument("--no-constraints", action="store_false", dest="constraints", help="ignore UCCA rules")
        group.add_argument("--max-nodes", type=float, default=3.0, help="max non-terminal/terminal ratio")
        group.add_argument("--max-height", type=int, default=20, help="max graph height")
        group = argparser.add_mutually_exclusive_group()
        group.add_argument("--no-swap", action="store_false", dest="swap", help="disable Swap transitions entirely")
        group.add_argument("--compound-swap", action="store_true", help="enable compound swap")
        group = argparser.add_argument_group(title="Sanity checks")
        group.add_argument("--check-loops", action="store_true", help="abort if the parser enters a state loop")
        group.add_argument("--verify", action="store_true", help="verify oracle reproduces original passage")
        group = group.add_mutually_exclusive_group()
        group.add_argument("-b", "--binary", action="store_true", help="read and write passages in Pickle")
        group.add_argument("-f", "--format", choices=convert.CONVERTERS, help="output format for parsed files")
        group = argparser.add_argument_group(title="General classifier training parameters")
        group.add_argument("--swap-importance", type=int, default=1, help="learning rate factor for Swap")
        group.add_argument("--early-update", action="store_true", help="move to next example on incorrect prediction")
        group.add_argument("--word-dim-external", type=int, default=300, help="dimension for external word embeddings")
        group.add_argument("--word-vectors", help="file to load external word embeddings from (default: GloVe)")
        group = argparser.add_argument_group(title="Perceptron parameters")
        group.add_argument("--learning-rate", type=float, default=1.0, help="rate for model weight updates")
        group.add_argument("--learning-rate-decay", type=float, default=0.0, help="learning rate decay per iteration")
        group.add_argument("--min-update", type=int, default=5, help="minimum #updates for using a feature")
        group = argparser.add_argument_group(title="Neural network parameters")
        group.add_argument("--update-word-vectors", action="store_true", help="tune the external word embeddings")
        group.add_argument("--word-dim", type=int, default=100, help="dimension for learned word embeddings")
        group.add_argument("--tag-dim", type=int, default=10, help="dimension for POS tag embeddings")
        group.add_argument("--dep-dim", type=int, default=10, help="dimension for dependency relation embeddings")
        group.add_argument("--label-dim", type=int, default=20, help="dimension for edge label embeddings")
        group.add_argument("--punct-dim", type=int, default=2, help="dimension for separator punctuation embeddings")
        group.add_argument("--gap-dim", type=int, default=2, help="dimension for gap type embeddings")
        group.add_argument("--action-dim", type=int, default=5, help="dimension for action type embeddings")
        group.add_argument("--layer-dim", type=int, default=500, help="dimension for hidden layers")
        group.add_argument("--layers", type=int, default=2, help="number of hidden layers")
        group.add_argument("--lstm-layer-dim", type=int, default=500, help="dimension for LSTM hidden layers")
        group.add_argument("--lstm-layers", type=int, default=2, help="number of LSTM hidden layers")
        group.add_argument("--embedding-layer-dim", type=int, default=500, help="dimension for layers before LSTM")
        group.add_argument("--embedding-layers", type=int, default=1, help="number of layers before LSTM")
        group.add_argument("--activation", choices=ACTIVATIONS, default=ACTIVATIONS[0], help="activation function")
        group.add_argument("--init", choices=INITIALIZATIONS, default=INITIALIZATIONS[0], help="weight initialization")
        group.add_argument("--max-labels", type=int, default=100, help="max number of actions to allow")
        group.add_argument("--save-every", type=int, help="every this many passages, evaluate on dev and save model")
        group.add_argument("--minibatch-size", type=int, default=200, help="mini-batch size for optimization")
        group.add_argument("--optimizer", choices=OPTIMIZERS, default=OPTIMIZERS[0], help="algorithm for optimization")
        group.add_argument("--max-words-external", type=int, help="max external word vectors to use")
        group.add_argument("--max-words", type=int, default=10000, help="max number of words to keep embeddings for")
        group.add_argument("--max-tags", type=int, default=100, help="max number of POS tags to keep embeddings for")
        group.add_argument("--max-deps", type=int, default=100, help="max number of dep labels to keep embeddings for")
        group.add_argument("--max-edge-labels", type=int, default=15, help="max number of edge labels for embeddings")
        group.add_argument("--max-puncts", type=int, default=5, help="max number of punctuations for embeddings")
        group.add_argument("--max-gaps", type=int, default=3, help="max number of gap types to keep embeddings for")
        group.add_argument("--max-actions", type=int, default=10, help="max number of action types for embeddings")
        group.add_argument("--word-dropout", type=float, default=0.25, help="word dropout parameter")
        group.add_argument("--word-dropout-external", type=float, default=0.25, help="word dropout for word vectors")
        group.add_argument("--dropout", type=float, default=0.5, help="dropout parameter between layers")
        group = argparser.add_argument_group(title="DyNet parameters")
        group.add_argument("--dynet-mem", help="memory for dynet")
        group.add_argument("--dynet-weight-decay", type=float, help="weight decay for parameters (default 1e-6)")
        group.add_argument("--dynet-gpu", action="store_true", help="use the GPU")
        group.add_argument("--dynet-gpus", type=int, help="how many GPUs you want to use")
        group.add_argument("--dynet-gpu-ids", help="the GPUs that you want to use by device ID")
        group.add_argument("--dynet-viz", action="store_true", help="visualize NN and exit")
        self.args = argparser.parse_args(args if args else None)
        
        if self.args.model:
            if not self.args.log:
                self.args.log = self.args.model + ".log"
            if self.args.dev and not self.args.devscores:
                self.args.devscores = self.args.model + ".dev.csv"
            if self.args.passages and not self.args.testscores:
                self.args.testscores = self.args.model + ".test.csv"
        elif not self.args.log:
            self.args.log = "parse.log"

        self._log_file = None
        self.set_external()
        self.random = np.random

    def set_external(self):
        np.random.seed(self.args.seed)
        sys.argv += ["--dynet-seed", str(self.args.seed)]
        if self.args.dynet_mem:
            sys.argv += ["--dynet-mem", str(self.args.dynet_mem)]
        if self.args.dynet_weight_decay:
            sys.argv += ["--dynet-weight-decay", str(self.args.dynet_weight_decay)]
        if self.args.dynet_gpu:
            sys.argv += ["--dynet-gpu"]
        if self.args.dynet_gpus:
            sys.argv += ["--dynet-gpus", str(self.args.dynet_gpus)]
        if self.args.dynet_gpu_ids:
            sys.argv += ["--dynet-gpu-ids", str(self.args.dynet_gpu_ids)]
        if self.args.dynet_viz:
            sys.argv += ["--dynet-viz"]

    def update(self, params):
        for name, value in params.items():
            setattr(self.args, name, value)
        self.set_external()

    @property
    def line_end(self):
        return "\n" if self.args.verbose else " "  # show all in one line unless verbose

    def log(self, message):
        try:
            if self._log_file is None:
                self._log_file = open(self.args.log, "w")
            print(message, file=self._log_file, flush=True)
            if self.args.verbose:
                print(message)
        except OSError:
            pass

    def close(self):
        if self._log_file is not None:
            self._log_file.close()

    def __str__(self):
        return " ".join("%s=%s" % item for item in sorted(vars(self.args).items()))



import numpy as np

from features.feature_extractor_wrapper import FeatureExtractorWrapper
from features.feature_params import MISSING_VALUE
from features.feature_params import NumericFeatureParameters
from tupa.model_util import DropoutDict


class FeatureEnumerator(FeatureExtractorWrapper):
    """
    Wrapper for DenseFeatureExtractor to replace non-numeric feature values with numbers.
    To be used with NeuralNetwork classifier.
    """

    def __init__(self, feature_extractor, params):
        super(FeatureEnumerator, self).__init__(feature_extractor, params)

    def init_params(self, param_list):
        params = super(FeatureEnumerator, self).init_params(param_list)
        for param in list(params.values()):
            if self.feature_extractor.features_exist(param.effective_suffix):
                self.init_data(param)
            else:
                del params[param.suffix]
        return params

    def init_data(self, param):
        if param.data is not None or isinstance(param, NumericFeatureParameters):
            return
        param.num = self.feature_extractor.num_features_non_numeric(param.effective_suffix)
        keys = ()
        if param.dim and param.external:
            vectors = self.get_word_vectors(param)
            keys = vectors.keys()
            param.init = np.array(list(vectors.values()))
        param.data = DropoutDict(max_size=param.size, keys=keys, dropout=param.dropout)

    def init_features(self, state, suffix):
        param = self.params[suffix]
        assert param.indexed, "Cannot initialize non-indexed parameter '%s'" % suffix
        values = self.feature_extractor.init_features(state, param.effective_suffix)
        assert MISSING_VALUE not in values, "Missing value occurred in feature initialization: '%s'" % suffix
        values = [param.data[v] for v in values]
        assert all(isinstance(f, int) for f in values), "Invalid feature numbers for '%s': %s" % (suffix, values)
        return values

    def extract_features(self, state):
        """
        Calculate feature values according to current state
        :param state: current state of the parser
        :return dict of feature name -> numeric value
        """
        numeric_features, non_numeric_features = self.feature_extractor.extract_features(state, self.params)
        features = {NumericFeatureParameters.SUFFIX: numeric_features}
        for suffix, param in self.params.items():
            if param.dim and (param.copy_from is None or not self.params[param.copy_from].dim or not param.indexed):
                values = non_numeric_features.get(param.effective_suffix)
                if values is not None:
                    features[suffix] = values if param.indexed else \
                        [v if v == MISSING_VALUE else param.data[v] for v in values]
                    # assert all(isinstance(f, int) for f in features[suffix]),\
                    #     "Invalid feature numbers for '%s': %s" % (suffix, features[suffix])
        return features

    def collapse_features(self, suffixes):
        self.feature_extractor.collapse_features({self.params[s].copy_from if self.params[s].external else s
                                                  for s in suffixes if self.params[s].dim})

    def filename_suffix(self):
        return "_enum"



import sys
from collections import deque, defaultdict

from tupa.action import Actions
from tupa.config import Config
from tupa.constants import Constraints
from states.edge import Edge
from states.node import Node
from ucca import core, layer0, layer1, textutil
from ucca.layer1 import EdgeTags


class State(object):
    """
    The parser's state, responsible for applying actions and creating the final Passage
    :param passage: a Passage object to get the tokens from, and everything else if training
    """
    def __init__(self, passage):
        self.log = []
        self.finished = False
        l0 = passage.layer(layer0.LAYER_ID)
        l1 = passage.layer(layer1.LAYER_ID)
        assert l0.all, "Empty passage '%s'" % passage.ID
        self.labeled = len(l1.all) > 1
        self.terminals = [Node(i, orig_node=t, text=t.text, paragraph=t.paragraph, tag=t.tag,
                               pos_tag=t.extra.get(textutil.TAG_KEY),
                               dep_rel=t.extra.get(textutil.DEP_KEY),
                               dep_head=t.extra.get(textutil.HEAD_KEY))
                          for i, t in enumerate(l0.all)]
        self.nodes = list(self.terminals)  # Copy the list of terminals; more nodes will be added later
        self.buffer = deque(self.nodes)  # Copy the list of nodes to initialize the buffer
        self.root = self.add_node(l1.heads[0])  # The root is not part of the buffer
        self.stack = [self.root]
        self.passage_id = passage.ID
        self.actions = []  # History of applied actions

    def is_valid(self, action):
        """
        :param action: action to check for validity
        :return: is the action (including tag) valid in the current state?
        """
        try:
            self.assert_valid(action)
        except AssertionError:
            return False
        return True

    def assert_valid(self, action):
        """
        Raise AssertionError if the action is invalid in the current state
        :param action: action to check for validity
        """
        def assert_possible_node():
            if self.labeled:  # We're in training, so we must have an original node to refer to
                assert action.orig_node is not None, "May only create real nodes during training"
            self.assert_node_ratio(extra=1)
            self.assert_height()

        def assert_possible_parent(node):
            assert node.text is None, "Terminals may not have children: %s" % node.text
            assert not node.implicit, "Implicit nodes may not have children: %s" % s0
            if Config().args.constraints:
                assert action.tag not in Constraints.UniqueOutgoing or action.tag not in node.outgoing_tags, \
                    "Outgoing edge tag %s must be unique, but %s already has one" % (
                        action.tag, node)
                assert action.tag not in Constraints.MutuallyExclusiveOutgoing or not \
                    node.outgoing_tags & Constraints.MutuallyExclusiveOutgoing, \
                    "Outgoing edge tags %s are mutually exclusive, but %s already has %s and is being added %s" % (
                        Constraints.MutuallyExclusiveOutgoing, node, node.outgoing_tags, action.tag)
                assert action.tag in Constraints.ChildlessOutgoing or not \
                    node.incoming_tags & Constraints.ChildlessIncoming, \
                    "Units with incoming %s edges may not have children, and %s has incoming %s" % (
                        Constraints.ChildlessIncoming, node, node.incoming_tags)

        def assert_possible_child(node):
            assert node is not self.root, "The root may not have parents"
            assert (node.text is not None) == (action.tag == EdgeTags.Terminal), \
                "Edge tag must be %s iff child is terminal, but node is %s and edge tag is %s" % (
                    EdgeTags.Terminal, node, action.tag)
            if Config().args.constraints:
                assert action.tag not in Constraints.UniqueIncoming or \
                    action.tag not in node.incoming_tags, \
                    "Incoming edge tag %s must be unique, but %s already has one" % (
                        action.tag, node)
                assert action.tag not in Constraints.ChildlessIncoming or \
                    node.outgoing_tags <= Constraints.ChildlessOutgoing, \
                    "Units with incoming %s edges may not have children, but %s has %d" % (
                        Constraints.ChildlessIncoming, node, len(node.children))
                assert action.remote or action.tag in Constraints.possible_multiple_incoming() or \
                    all(e.remote or e.tag in Constraints.possible_multiple_incoming()
                        for e in node.incoming), \
                    "Multiple parents only allowed if they are remote or linkage edges: %s, %s" % (
                        action, node)
                # Commented out due to passage 106, unit 1.300
                # assert not node.incoming_tags or (action.tag in Constraints.LinkerIncoming) == (
                #     node.incoming_tags <= Constraints.LinkerIncoming), \
                #     "Linker units may only have incoming edges with tags from %s, but %s is being added '%s'" % (
                #         Constraints.LinkerIncoming, node, action.tag)

        def assert_possible_edge():
            parent, child = self.get_parent_child(action)
            assert_possible_parent(parent)
            assert_possible_child(child)
            if parent is self.root and Config().args.constraints:
                assert child.text is None, "Root may not have terminal children, but is being added '%s'" % child
                assert action.tag in Constraints.TopLevel, "The root may not have %s edges" % action.tag
            # if Config().args.multiple_edges:  # Removed this option because it is not useful right now
            #     edge = Edge(parent, child, action.tag, remote=action.remote)
            #     assert edge not in parent.outgoing, "Edge must not already exist: %s" % edge
            # else:
            assert child not in parent.children, "Edge must not already exist: %s->%s" % (parent, child)
            assert parent not in child.descendants, "Detected cycle created by edge: %s->%s" % (parent, child)

        if action.is_type(Actions.Finish):
            if Config().args.swap:  # Without swap, the oracle may be incapable even of single action
                assert self.root.outgoing, \
                    "Root must have at least one child at the end of the parse, but has none"
        elif action.is_type(Actions.Shift):
            assert self.buffer, "Buffer must not be empty in order to shift from it"
        else:  # Unary actions
            assert self.actions, "First action must be Shift, but was %s" % action
            assert self.stack, "Action requires non-empty stack: %s" % action
            s0 = self.stack[-1]
            if action.is_type(Actions.Node):
                assert_possible_child(s0)
                assert_possible_node()
            elif action.is_type(Actions.Implicit):
                assert_possible_parent(s0)
                assert_possible_node()
            elif action.is_type(Actions.Reduce):
                assert s0 is not self.root or s0.outgoing, "May not reduce the root without children"
                # Commented out due to passage 126, unit 1.338
                # assert not s0.outgoing_tags & Constraints.SceneSufficientOutgoing and \
                #     not s0.incoming_tags & Constraints.SceneSufficientIncoming or \
                #     s0.outgoing_tags & Constraints.SceneNecessaryOutgoing, \
                #     "May not reduce a scene before it has any outgoing edge of %s (it has only %s)" % (
                #         Constraints.SceneNecessaryOutgoing, s0.outgoing_tags)
                # Commented out due to passage 126, unit 1.60
                # assert s0.incoming_tags == Constraints.LinkerIncoming or not \
                #     s0.incoming_tags & Constraints.LinkerIncoming, \
                #     "May not reduce a linker before it has all incoming edges of %s (it has only %s)" % (
                #         Constraints.LinkerIncoming, s0.incoming_tags)
            else:  # Binary actions
                assert len(self.stack) > 1, "Action requires at least two stack elements: %s" % action
                if action.is_type(Actions.LeftEdge, Actions.RightEdge, Actions.LeftRemote, Actions.RightRemote):
                    assert_possible_edge()
                elif action.is_type(Actions.Swap):
                    # A regular swap is possible since the stack has at least two elements;
                    # A compound swap is possible if the stack is longer than the distance
                    distance = action.tag or 1
                    assert 1 <= distance < len(self.stack), "Invalid swap distance: %d" % distance
                    swapped = self.stack[-distance - 1]
                    # To prevent swap loops: only swap if the nodes are currently in their original order
                    assert self.swappable(s0, swapped),\
                        "Swapping already-swapped nodes: %s (swap index %d) <--> %s (swap index %d)" % (
                            swapped, swapped.swap_index, s0, s0.swap_index)
                else:
                    raise Exception("Invalid action: %s" % action)

    @staticmethod
    def swappable(right, left):
        return left.swap_index < right.swap_index

    # noinspection PyTypeChecker
    def transition(self, action):
        """
        Main part of the parser: apply action given by oracle or classifier
        :param action: Action object to apply
        """
        action.apply()
        self.log = []
        if action.is_type(Actions.Shift):  # Push buffer head to stack; shift buffer
            self.stack.append(self.buffer.popleft())
        elif action.is_type(Actions.Node):  # Create new parent node and add to the buffer
            parent = self.add_node(action.orig_node)
            self.update_swap_index(parent)
            self.add_edge(Edge(parent, self.stack[-1], action.tag))
            self.buffer.appendleft(parent)
        elif action.is_type(Actions.Implicit):  # Create new child node and add to the buffer
            child = self.add_node(action.orig_node, implicit=True)
            self.update_swap_index(child)
            self.add_edge(Edge(self.stack[-1], child, action.tag))
            self.buffer.appendleft(child)
        elif action.is_type(Actions.Reduce):  # Pop stack (no more edges to create with this node)
            self.stack.pop()
        elif action.is_type(Actions.LeftEdge, Actions.LeftRemote, Actions.RightEdge, Actions.RightRemote):
            parent, child = self.get_parent_child(action)
            self.add_edge(Edge(parent, child, action.tag, remote=action.remote))
        elif action.is_type(Actions.Swap):  # Place second (or more) stack item back on the buffer
            distance = action.tag or 1
            s = slice(-distance - 1, -1)
            self.log.append("%s <--> %s" % (", ".join(map(str, self.stack[s])), self.stack[-1]))
            self.buffer.extendleft(reversed(self.stack[s]))  # extendleft reverses the order
            del self.stack[s]
        elif action.is_type(Actions.Finish):  # Nothing left to do
            self.finished = True
        else:
            raise Exception("Invalid action: " + action)
        if Config().args.verify:
            intersection = set(self.stack).intersection(self.buffer)
            assert not intersection, "Stack and buffer overlap: %s" % intersection
        self.assert_node_ratio()
        action.index = len(self.actions)
        self.actions.append(action)

    def add_node(self, *args, **kwargs):
        """
        Called during parsing to add a new Node (not core.Node) to the temporary representation
        :param args: ordinal arguments for Node()
        :param kwargs: keyword arguments for Node()
        """
        node = Node(len(self.nodes), *args, **kwargs)
        if Config().args.verify:
            assert node not in self.nodes, "Node already exists"
        self.nodes.append(node)
        self.log.append("node: %s" % node)
        return node

    def add_edge(self, edge):
        edge.add()
        self.log.append("edge: %s" % edge)

    def get_parent_child(self, action):
        if action.is_type(Actions.LeftEdge, Actions.LeftRemote):
            return self.stack[-1], self.stack[-2]
        elif action.is_type(Actions.RightEdge, Actions.RightRemote):
            return self.stack[-2], self.stack[-1]
        else:
            return None, None

    def create_passage(self, assert_proper=True):
        """
        Create final passage from temporary representation
        :param assert_proper: fail if this results in an improper passage
        :return: core.Passage created from self.nodes
        """
        passage = core.Passage(self.passage_id)
        l0 = layer0.Layer0(passage)
        terminals = [l0.add_terminal(text=terminal.text, punct=terminal.tag == layer0.NodeTags.Punct,
                                     paragraph=terminal.paragraph) for terminal in self.terminals]
        l1 = layer1.Layer1(passage)
        if self.labeled:  # We are in training and we have a gold passage
            l1.heads[0].extra["remarks"] = self.root.node_id  # For reference
            self.fix_terminal_tags(terminals)
        remotes = []  # To be handled after all nodes are created
        linkages = []  # To be handled after all non-linkage nodes are created
        self.topological_sort()  # Sort self.nodes
        for node in self.nodes:
            if self.labeled and assert_proper:
                assert node.text or node.outgoing or node.implicit, "Non-terminal leaf node: %s" % node
                assert node.node or node is self.root or node.is_linkage, "Non-root without incoming: %s" % node
            if node.is_linkage:
                linkages.append(node)
            else:
                for edge in node.outgoing:
                    if edge.remote:
                        remotes.append((node, edge))
                    else:
                        edge.child.add_to_l1(l1, node, edge.tag, terminals, self.labeled)

        for node, edge in remotes:  # Add remote edges
            try:
                assert node.node is not None, "Remote edge from nonexistent node"
                assert edge.child.node is not None, "Remote edge to nonexistent node"
                node.node.add(edge.tag, edge.child.node, edge_attrib={"remote": True})
            except AssertionError:
                if assert_proper:
                    raise

        for node in linkages:  # Add linkage nodes and edges
            try:
                link_relation = None
                link_args = []
                for edge in node.outgoing:
                    assert edge.child.node is not None, "Linkage edge to nonexistent node"
                    if edge.tag == EdgeTags.LinkRelation:
                        assert link_relation is None, \
                            "Multiple link relations: %s, %s" % (link_relation, edge.child.node)
                        link_relation = edge.child.node
                    elif edge.tag == EdgeTags.LinkArgument:
                        link_args.append(edge.child.node)
                assert link_relation is not None, "No link relations: %s" % node
                # if len(link_args) < 2:
                #     print("Less than two link arguments for linkage %s" % node, file=sys.stderr)
                node.node = l1.add_linkage(link_relation, *link_args)
                if node.node_id:  # We are in training and we have a gold passage
                    node.node.extra["remarks"] = node.node_id  # For reference
            except AssertionError:
                if assert_proper:
                    raise

        return passage

    def fix_terminal_tags(self, terminals):
        for terminal, orig_terminal in zip(terminals, self.terminals):
            if terminal.tag != orig_terminal.tag:
                if Config().args.verbose:
                    print("%s is the wrong tag for terminal: %s" % (terminal.tag, terminal.text),
                          file=sys.stderr)
                terminal.tag = orig_terminal.tag

    def topological_sort(self):
        """
        Sort self.nodes topologically, each node appearing as early as possible
        Also sort each node's outgoing and incoming edge according to the node order
        """
        levels = defaultdict(list)
        level_by_index = {}
        stack = [node for node in self.nodes if not node.outgoing]
        while stack:
            node = stack.pop()
            if node.index not in level_by_index:
                parents = [edge.parent for edge in node.incoming]
                if parents:
                    unexplored_parents = [parent for parent in parents
                                          if parent.index not in level_by_index]
                    if unexplored_parents:
                        for parent in unexplored_parents:
                            stack.append(node)
                            stack.append(parent)
                    else:
                        level = 1 + max(level_by_index[parent.index] for parent in parents)
                        levels[level].append(node)
                        level_by_index[node.index] = level
                else:
                    levels[0].append(node)
                    level_by_index[node.index] = 0
        self.nodes = [node for level, level_nodes in sorted(levels.items())
                      for node in sorted(level_nodes, key=lambda x: x.node_index or x.index)]
        for node in self.nodes:
            node.outgoing.sort(key=lambda x: x.child.node_index or self.nodes.index(x.child))
            node.incoming.sort(key=lambda x: x.parent.node_index or self.nodes.index(x.parent))

    def node_ratio(self, extra=0):
        return (len(self.nodes) + extra) / len(self.terminals) - 1

    def assert_node_ratio(self, extra=0):
        max_ratio = Config().args.max_nodes
        assert self.node_ratio(extra=extra) <= max_ratio, \
            "Reached maximum ratio (%.3f) of non-terminals to terminals" % max_ratio

    def assert_height(self):
        max_height = Config().args.max_height
        assert self.root.height <= max_height, \
            "Reached maximum graph height (%d)" % max_height

    def update_swap_index(self, node):
        """
        Update the node's swap index according to the nodes before and after it.
        Usually the swap index is just the index, and that is what it is initialized to.
        If the buffer is not empty and the next node on it is not a terminal, it means that it is
        a non-terminal that was created at some point, probably before this node (because this method
        should be run just when this node is created).
        In that case, the buffer head's index will be lower than this node's index, and we will
        update this node's swap index to be the arithmetic average between the previous node
        (stack top) and the next node (buffer head).
        This will make sure that when we perform the validity check on the SWAP action,
        we will correctly identify this node as always having appearing before b (what is the
        current buffer head). Otherwise, we would prevent swapping this node with b even though
        it should be a legal action (because they have never been swapped before).
        :param node: the new node that was added
        """
        if self.buffer:
            b = self.buffer[0]
            if self.stack and (b.text is not None or b.swap_index <= node.swap_index):
                s = self.stack[-1]
                node.swap_index = (s.swap_index + b.swap_index) / 2

    def str(self, sep):
        return "stack: [%-20s]%sbuffer: [%s]" % (" ".join(map(str, self.stack)), sep,
                                                 " ".join(map(str, self.buffer)))

    def __str__(self):
        return self.str(" ")

    def __eq__(self, other):
        return self.stack == other.stack and self.buffer == other.buffer and \
               self.nodes == other.nodes

    def __hash__(self):
        return hash((tuple(self.stack), tuple(self.buffer), tuple(self.nodes)))



import pickle
import time

import numpy as np
import os
from collections import defaultdict, Counter

from features.feature_params import UNKNOWN_VALUE


class UnknownDict(defaultdict):
    """
    defaultdict that has a single default value for missing keys
    """
    UNKNOWN = "<UNKNOWN>"

    def __init__(self, d, unknown=None):
        """
        :param d: base dict to initialize by
        :param unknown: value to return for missing keys
        """
        # noinspection PyTypeChecker
        super(UnknownDict, self).__init__(None, d)
        if self.UNKNOWN not in self:
            assert unknown is not None, "Default value must not be None"
            self[self.UNKNOWN] = unknown
        self.unknown = self[self.UNKNOWN]

    def __missing__(self, key):
        return self.unknown


class AutoIncrementDict(UnknownDict):
    """
    defaultdict that returns an auto-incrementing index for new keys
    """
    def __init__(self, max_size=None, keys=(), d=None):
        """
        :param max_size: maximum index to allow, after which `unknown' will be returned
        :param keys: initial sequence of keys
        :param d: dictionary to initialize from
        """
        super(AutoIncrementDict, self).__init__({} if d is None else d, unknown=UNKNOWN_VALUE)
        self.max = max_size
        for key in keys:
            self.__missing__(key)

    def __missing__(self, key):
        ret = self[key] = len(self) if self.max is not None and len(self) < self.max else self.unknown
        return ret


class DropoutDict(AutoIncrementDict):
    """
    UnknownDict that sometimes returns the unknown value even for existing keys
    """
    def __init__(self, d=None, dropout=0, max_size=None, keys=()):
        """
        :param d: base dict to initialize by
        :param dropout: dropout parameter
        """
        super(DropoutDict, self).__init__(max_size, keys, d=d)
        assert dropout >= 0, "Dropout value must be >= 0, but given %f" % dropout
        if d is not None and isinstance(d, DropoutDict):
            self.dropout = d.dropout
            self.counts = d.counts if self.dropout > 0 else None
        else:
            self.dropout = dropout
            self.counts = Counter() if self.dropout > 0 else None

    def __getitem__(self, item):
        if item != self.UNKNOWN and self.dropout > 0:
            self.counts[item] += 1
            if self.dropout / (self.counts[item] + self.dropout) > np.random.random_sample():
                item = UnknownDict.UNKNOWN
        return super(DropoutDict, self).__getitem__(item)


def save_dict(filename, d):
    """
    Save dictionary to file
    :param filename: file to write to
    :param d: dictionary to save
    """
    print("Saving to '%s'... " % filename, end="", flush=True)
    started = time.time()
    with open(filename, 'wb') as h:
        pickle.dump(d, h, protocol=pickle.HIGHEST_PROTOCOL)
    print("Done (%.3fs)." % (time.time() - started))


def load_dict(filename):
    """
    Load dictionary from file
    :param filename: file to read from
    """

    def try_load(*names):
        exception = None
        for f in names:
            # noinspection PyBroadException
            try:
                with open(f, 'rb') as h:
                    return pickle.load(h)
            except Exception as e:
                exception = e
        if exception is not None:
            raise FileNotFoundError("File not found: '%s'" % filename) from exception

    print("Loading from '%s'... " % filename, end="", flush=True)
    started = time.time()
    d = try_load(filename, os.path.splitext(filename)[0])
    print("Done (%.3fs)." % (time.time() - started))
    return d



from copy import copy

MISSING_VALUE = -1
UNKNOWN_VALUE = 0


class FeatureParameters(object):
    def __init__(self, suffix, dim, size, dropout=0, updated=True, num=1, init=None, data=None, indexed=False,
                 copy_from=None, filename=None):
        """
        :param suffix: one-character title for feature
        :param dim: vector dimension or, filename to load vectors from, or Word2Vec object
        :param size: maximum number of distinct values
        :param dropout: value of dropout parameter to use during training
        :param updated: whether the feature is learned (otherwise kept constant)
        :param num: how many such features exist per step
        :param init: array of values to use as initial value of embedding matrix
        :param data: dictionary of raw value to running numerical representation, or embedding matrix
        :param indexed: whether the feature is to be used as index into initialized values (otherwise used directly)
        :param copy_from: suffix of other parameter to copy values from instead of extracting them directly
        :param filename: name of file to load data from
        """
        self.suffix = suffix
        self.dim = dim
        self.size = size
        self.dropout = dropout
        self.updated = updated
        self.num = num
        self.init = init
        self.data = data
        self.indexed = indexed
        self.copy_from = copy_from
        self.filename = filename

    def __repr__(self):
        return "%s(%s, %d, %d, %f, %s, %s, %s, %s, %s, %s, %s)" % (
            self.__class__.__name__, self.suffix, self.dim, self.size, self.dropout, self.updated, self.num, self.init,
            self.data, self.indexed, self.copy_from, self.filename)

    @property
    def numeric(self):
        return False

    @property
    def effective_suffix(self):
        return self.suffix if self.copy_from is None else self.copy_from

    @property
    def external(self):
        return self.copy_from is not None


class NumericFeatureParameters(FeatureParameters):
    SUFFIX = "numeric"

    def __init__(self, num):
        super(NumericFeatureParameters, self).__init__(NumericFeatureParameters.SUFFIX, 1, None, num=num)

    def __repr__(self):
        return "%s(%d)" % (
            self.__class__.__name__, self.num)

    @property
    def numeric(self):
        return True


def copy_params(params, copy_dict=dict):
    params_copy = {}
    for suffix, param in params.items():
        param_copy = copy(param)
        if param.data is not None:
            param_copy.data = copy_dict(param.data)
        params_copy[suffix] = param_copy
    return params_copy



from features.feature_extractor_wrapper import FeatureExtractorWrapper

INDEXED_FEATURES = "W", "w", "t", "d"  # external word embeddings, learned word embeddings, POS tags, dep rels


class FeatureIndexer(FeatureExtractorWrapper):
    """
    Wrapper for FeatureEnumerator to replace non-numeric feature values with indices.
    To be used with BiLSTM classifier.
    """
    def __init__(self, feature_extractor, params=None):
        super(FeatureIndexer, self).__init__(feature_extractor, feature_extractor.params if params is None else params)
        if params is None:
            for suffix in INDEXED_FEATURES:
                param = self.params.get(suffix)
                if param is not None:
                    param.indexed = True
        else:
            feature_extractor.params = params
        self.feature_extractor.collapse_features(INDEXED_FEATURES)



import re

from tupa.config import Config, Singleton


class Action(object):
    type_to_id = {}

    RIGHT = 0
    LEFT = 1
    MAX_SWAP = 15  # default maximum size for compound swap

    def __init__(self, action_type, tag=None, orig_edge=None, orig_node=None, oracle=None):
        self.type = action_type  # String
        self.tag = tag  # Usually the tag of the created edge; but if COMPOUND_SWAP, the distance
        self.orig_node = orig_node  # Node created by this action, if any (during training)
        self.orig_edge = orig_edge
        self.edge = None  # Will be set by State when the edge created by this action is known
        self.oracle = oracle
        self.index = None

        self.type_id = Action.type_to_id.get(self.type)  # Allocate ID for fast comparison
        if self.type_id is None:
            self.type_id = len(Action.type_to_id)
            Action.type_to_id[self.type] = self.type_id
        self._id = None

    def is_type(self, *others):
        return self.type_id in (o.type_id for o in others)

    def apply(self):
        if self.oracle is not None:
            self.oracle.remove(self.orig_edge, self.orig_node)

    @staticmethod
    def from_string(s):
        m = re.match("(.*)-(.*)", s)
        if m:  # String contains tag
            action_type, tag = m.groups()
            return Action(action_type, tag)
        return Action(s)

    @classmethod
    def edge_action(cls, direction, remote, tag, *args, **kwargs):
        remote_action, edge_action = (Actions.RightRemote, Actions.RightEdge) \
            if direction == cls.RIGHT else (Actions.LeftRemote, Actions.LeftEdge)
        action = remote_action if remote else edge_action
        return action(tag, *args, **kwargs)

    def __repr__(self):
        return Action.__name__ + "(" + self.type + (", " + self.tag if self.tag else "") + ")"

    def __str__(self):
        return self.type + ("-" + str(self.tag) if self.tag else "")

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __call__(self, *args, **kwargs):
        return Action(self.type, *args, **kwargs)

    @property
    def remote(self):
        return self.is_type(Actions.LeftRemote, Actions.RightRemote)

    @property
    def is_swap(self):
        return self.is_type(Actions.Swap)

    @property
    def id(self):
        self.generate_id()
        return self._id

    def generate_id(self):
        if self._id is None:
            key = (self.type_id, self.tag)
            actions = Actions()
            self._id = actions.ids.get(key)
            if self._id is None:  # New action, add to list
                # noinspection PyTypeChecker
                self._id = len(actions.all)
                actions.all.append(self)
                actions.ids[key] = self._id


class Actions(object, metaclass=Singleton):

    Shift = Action("SHIFT")
    Node = Action("NODE")
    Implicit = Action("IMPLICIT")
    Reduce = Action("REDUCE")
    LeftEdge = Action("LEFT-EDGE")
    RightEdge = Action("RIGHT-EDGE")
    LeftRemote = Action("LEFT-REMOTE")
    RightRemote = Action("RIGHT-REMOTE")
    Swap = Action("SWAP")
    Finish = Action("FINISH")

    def __init__(self):
        self._all = None
        self._ids = None

    def init(self):
        # edge and node action will be created as they are returned by the oracle
        self.all = [Actions.Reduce, Actions.Shift, Actions.Finish] + \
                   (map(Actions.Swap, range(1, Action.MAX_SWAP)) if Config().args.compound_swap
                    else [Actions.Swap] if Config().args.swap else [])

    @property
    def all(self):
        if self._all is None:
            self.init()
        return self._all

    @all.setter
    def all(self, actions):
        self._all = list(actions)
        self._ids = {(action.type_id, action.tag): i for i, action in enumerate(actions)}

    @property
    def ids(self):
        if self._all is None:
            self.init()
        return self._ids





import re

from ucca import layer0
from ucca.layer1 import EdgeTags

FEATURE_ELEMENT_PATTERN = re.compile("([sba])(\d)([lruLRU]*)([wtdhepqxyAPCIR]*)")
FEATURE_TEMPLATE_PATTERN = re.compile("^(%s)+$" % FEATURE_ELEMENT_PATTERN.pattern)


class FeatureTemplate(object):
    """
    A feature template in parsed form, ready to be used for value calculation
    """
    def __init__(self, name, elements):
        """
        :param name: name of the feature in the short-hand form, to be used for the dictionary
        :param elements: collection of FeatureElement objects that represent the actual feature
        """
        self.name = name
        self.suffix = name[-1]
        self.elements = elements

    def __str__(self):
        return self.name


class FeatureTemplateElement(object):
    """
    One element in the values of a feature, e.g. from one node
    """
    def __init__(self, source, index, relatives, properties):
        """
        :param source: where to take the data from:
                           s: stack nodes
                           b: buffer nodes
                           a: past actions
        :param index: non-negative integer, the index of the element in the stack, buffer or list
                           of past actions (in the case of stack and actions, indexing from the end)
        :param relatives: string in [lruLRU]*, to select a descendant/parent of the node instead:
                           l: leftmost child
                           r: rightmost child
                           u: only child, if there is just one
                           L: leftmost parent
                           R: rightmost parent
                           U: only parent, if there is just one
        :param properties: the actual values to choose, if available (else omit feature), out of:
                           w: node text
                           t: node POS tag
                           d: node dependency relation
                           h: node height
                           e: tag of first incoming edge / action tag
                           ,: unique separator punctuation between nodes
                           q: count of any separator punctuation between nodes
                           x: gap type
                           y: sum of gap lengths
                           A: action type label
                           P: number of parents
                           C: number of children
                           I: number of implicit children
                           R: number of remote children
                           If empty, the value will be 1 if there is an edge from this node to the
                           next one in the template, or 0 otherwise. Also, if the next node comes
                           with the "e" property, then the edge with this node will be considered.
        """
        self.source = source
        self.index = int(index)
        self.relatives = relatives
        self.properties = properties

    def __str__(self):
        return self.source + str(self.index) + self.relatives + self.properties

    def __eq__(self, other):
        return self.source == other.source and self.index == other.index and self.relatives == other.relatives


class FeatureExtractor(object):
    """
    Object to extract features from the parser state to be used in action classification
    """
    def __init__(self, feature_templates):
        assert all(FEATURE_TEMPLATE_PATTERN.match(f) for f in feature_templates), \
            "Features do not match pattern: " + ", ".join(
                f for f in feature_templates if not FEATURE_TEMPLATE_PATTERN.match(f))
        # convert the list of features textual descriptions to the actual fields
        self.feature_templates = [FeatureTemplate(
            feature_name, tuple(FeatureTemplateElement(*m.group(1, 2, 3, 4))
                                for m in re.finditer(FEATURE_ELEMENT_PATTERN, feature_name)))
                                  for feature_name in feature_templates]

    def extract_features(self, state, params):
        """
        Calculate feature values according to current state
        :param state: current state of the parser
        :param params: dict of FeatureParameters for each suffix
        """
        raise NotImplementedError()

    def init_features(self, state, suffix=None):
        """
        Calculate feature values for initial state
        :param state: initial state of the parser
        :param suffix: feature suffix to get
        """
        pass

    def collapse_features(self, suffixes):
        """
        For each set of features referring to the same node, with the given properties,
         leave only one of them.
        """
        pass

    def finalize(self):
        return self

    def save(self, filename):
        pass

    def load(self, filename):
        pass

    @staticmethod
    def calc_feature(feature_template, state, default=None, indexed=False):
        values = []
        prev_elem = None
        prev_node = None
        for element in feature_template.elements:
            node = FeatureExtractor.get_node(element, state)
            if not element.properties:
                if prev_elem is not None:
                    if node is None or prev_node is None:
                        if default is None:
                            return None
                        values.append(default)
                    else:
                        values.append(int(prev_node in node.parents))
                prev_elem = element
                prev_node = node
            else:
                prev_elem = None
                prev_node = None
                for p in ("i",) if indexed else element.properties:
                    v = FeatureExtractor.get_prop(element, node, prev_node, p, state)
                    if v is None:
                        if default is None:
                            return None
                        values.append(default)
                    else:
                        values.append(v)
        return values

    @staticmethod
    def get_prop(element, node, prev_node, p, state):
        try:
            if element is not None and element.source == "a":
                return FeatureExtractor.get_action_prop(node, p)
            elif p in "pq":
                return FeatureExtractor.get_separator_prop(
                    state.stack[-1:-3:-1], state.terminals, p)
            else:
                return FeatureExtractor.get_node_prop(node, p, prev_node)
        except (AttributeError, StopIteration):
            return None

    @staticmethod
    def get_node(element, state):
        if element.source == "s":
            if len(state.stack) <= element.index:
                return None
            node = state.stack[-1 - element.index]
        elif element.source == "b":
            if len(state.buffer) <= element.index:
                return None
            node = state.buffer[element.index]
        else:  # source == "a"
            if len(state.actions) <= element.index:
                return None
            node = state.actions[-1 - element.index]
        for relative in element.relatives:
            nodes = node.parents if relative.isupper() else node.children
            lower = relative.lower()
            if not nodes:
                return None
            elif len(nodes) == 1:
                if lower == "u":
                    node = nodes[0]
            elif lower == "l":
                node = nodes[0]
            elif lower == "r":
                node = nodes[-1]
            else:  # lower == "u" and len(nodes) > 1
                return None
        return node

    NODE_PROP_GETTERS = {
        "w": lambda node, _: FeatureExtractor.get_head_terminal(node).text,
        "t": lambda node, _: FeatureExtractor.get_head_terminal(node).pos_tag,
        "d": lambda node, _: FeatureExtractor.get_head_terminal(node).dep_rel,
        "h": lambda node, _: FeatureExtractor.get_height(node),
        "i": lambda node, _: FeatureExtractor.get_head_terminal(node).index,
        "e": lambda node, prev_node: next(e.tag for e in node.incoming
                                          if prev_node is None or e.parent == prev_node),
        "x": lambda node, _: FeatureExtractor.gap_type(node),
        "y": lambda node, _: FeatureExtractor.gap_length_sum(node),
        "P": lambda node, _: len(node.incoming),
        "C": lambda node, _: len(node.outgoing),
        "I": lambda node, _: len([n for n in node.children if n.implicit]),
        "R": lambda node, _: len([e for e in node.outgoing if e.remote]),
    }

    @staticmethod
    def get_node_prop(node, p, prev_node=None):
        return FeatureExtractor.NODE_PROP_GETTERS[p](node, prev_node)

    ACTION_PROPS = {
        "A": "type",
        "e": "tag",
    }

    @staticmethod
    def get_action_prop(action, p):
        return getattr(action, FeatureExtractor.ACTION_PROPS[p])

    @staticmethod
    def get_separator_prop(nodes, terminals, p):
        if len(nodes) < 2:
            return None
        t0, t1 = sorted([FeatureExtractor.get_head_terminal(node) for node in nodes],
                        key=lambda t: t.index)
        punctuation = [terminal for terminal in terminals[t0.index + 1:t1.index]
                       if terminal.tag == layer0.NodeTags.Punct]
        if p == "p" and len(punctuation) == 1:
            return punctuation[0].text
        if p == "q":
            return len(punctuation)
        return None

    EDGE_PRIORITY = {tag: i for i, tag in enumerate((
        EdgeTags.Center,
        EdgeTags.Connector,
        EdgeTags.ParallelScene,
        EdgeTags.Process,
        EdgeTags.State,
        EdgeTags.Participant,
        EdgeTags.Adverbial,
        EdgeTags.Time,
        EdgeTags.Elaborator,
        EdgeTags.Relator,
        EdgeTags.Function,
        EdgeTags.Linker,
        EdgeTags.LinkRelation,
        EdgeTags.LinkArgument,
        EdgeTags.Ground,
        EdgeTags.Terminal,
        EdgeTags.Punctuation,
    ))}

    @staticmethod
    def get_head_terminal(node):
        return FeatureExtractor.get_head_terminal_height(node)

    @staticmethod
    def get_height(node):
        return FeatureExtractor.get_head_terminal_height(node, True)

    @staticmethod
    def get_head_terminal_height(node, return_height=False):
        height = 0
        while node.text is None:  # Not a terminal
            edges = [edge for edge in node.outgoing
                     if not edge.remote and not edge.child.implicit]
            if not edges or height > 30:
                return None
            node = min(edges, key=lambda edge: FeatureExtractor.EDGE_PRIORITY.get(
                edge.tag, 0)).child
            height += 1
        return height if return_height else node

    @staticmethod
    def has_gaps(node):
        # Possibly the same as FoundationalNode.discontiguous
        return any(length > 0 for length in FeatureExtractor.gap_lengths(node))

    @staticmethod
    def gap_length_sum(node):
        return sum(FeatureExtractor.gap_lengths(node))

    @staticmethod
    def gap_lengths(node):
        terminals = node.get_terminals()
        return (t1.index - t2.index - 1 for (t1, t2) in zip(terminals[1:], terminals[:-1]))

    @staticmethod
    def gap_type(node):
        if node.text is not None:
            return "n"  # None
        if FeatureExtractor.has_gaps(node):
            return "p"  # Pass
        if any(child.text is None and FeatureExtractor.has_gaps(child)
               for child in node.children):
            return "s"  # Source
        return "n"  # None



from collections import defaultdict

import numpy as np

from linear.perceptron import Perceptron
from tupa.config import Config, SPARSE_PERCEPTRON


class FeatureWeights(object):
    """
    The weights for one feature, for all labels
    """
    def __init__(self, num_labels=None, weights=None):
        if num_labels is not None and weights is None:
            self.weights = np.zeros(num_labels, dtype=float)  # 0.01 * np.random.randn(num_labels)
            self._totals = np.zeros(num_labels, dtype=float)
            self._last_update = np.zeros(num_labels, dtype=int)
            self.update_count = 0
        else:
            self.weights = weights

    def update(self, label, value, update_index):
        """
        Add a value to the entry of the given label
        :param label: label to index by
        :param value: value to add
        :param update_index: which update this is (for averaging)
        """
        self.update_count += 1
        self._update_totals(label, update_index)
        self.weights[label] += value

    def finalize(self, update_index, average=True):
        """
        Average weights over all updates, and keep only true label columns
        :param update_index: number of updates to average over
        :param average: whether to really average the weights or just return them as they are now
        :return new Weights object with the weights averaged and only selected indices remaining
        """
        self._update_totals(None, update_index)
        weights = self._totals / update_index if average else self.weights
        return FeatureWeights(weights=weights)

    def _update_totals(self, label, update_index):
        self._totals[label] += self.weights[label] * (update_index - self._last_update[label])
        self._last_update[label] = update_index

    def resize(self, num_labels):
        self.weights.resize(num_labels, refcheck=False)
        self._totals.resize(num_labels, refcheck=False)
        self._last_update.resize(num_labels, refcheck=False)


class SparsePerceptron(Perceptron):
    """
    Multi-class averaged perceptron with min-update for sparse features.
    Keeps weights in a dictionary by feature name, allowing adding new features on-the-fly.
    Also allows adding new labels on-the-fly.
    Expects features from SparseFeatureExtractor.
    """

    def __init__(self, *args, model=None, epoch=0):
        """
        Create a new untrained Perceptron or copy the weights from an existing one
        :param labels: a list of labels that can be updated later to add a new label
        :param min_update: minimum number of updates to a feature required for consideration
        :param model: if given, copy the weights (from a trained model)
        """
        super(SparsePerceptron, self).__init__(SPARSE_PERCEPTRON, *args, model=model, epoch=epoch)
        model = defaultdict(lambda: FeatureWeights(self.num_labels))
        if self.is_frozen:
            model.update(self.model)
        self.model = model
        self._min_update = Config().args.min_update  # Minimum number of updates for a feature to be used in scoring

    def score(self, features):
        """
        Calculate score for each label
        :param features: extracted feature values, in the form of a dict (name -> value)
        :return: array with score for each label
        """
        super(SparsePerceptron, self).score(features)
        scores = np.zeros(self.num_labels)
        for feature, value in features.items():
            if not value:
                continue
            weights = self.model.get(feature)
            if weights is None or not self.is_frozen and weights.update_count < self._min_update:
                continue
            scores += value * weights.weights
        return scores

    def update(self, features, pred, true, importance=1):
        """
        Update classifier weights according to predicted and true labels
        :param features: extracted feature values, in the form of a dict (name: value)
        :param pred: label predicted by the classifier (non-negative integer less than num_labels)
        :param true: true label (non-negative integer less than num_labels)
        :param importance: how much to scale the feature vector for the weight update
        """
        super(SparsePerceptron, self).update(features, pred, true, importance)
        for feature, value in features.items():
            if not value:
                continue
            weights = self.model[feature]
            weights.update(true, importance * self.learning_rate * value, self._update_index)
            weights.update(pred, -importance * self.learning_rate * value, self._update_index)

    def resize(self):
        for weights in self.model.values():
            weights.resize(self.num_labels)
        self.model.default_factory = lambda: FeatureWeights(self.num_labels)

    def _finalize_model(self, average):
        model = {f: w.finalize(self._update_index, average=average)
                 for f, w in self.model.items() if w.update_count >= self._min_update}
        print("%d features occurred at least %d times" % (len(model), self._min_update))
        return SparsePerceptron(self.filename, list(self.labels), model=model, epoch=self.epoch)

    def save_extra(self):
        return {
            "model": dict(self.model),
            "_min_update": self._min_update,
        }

    def load_extra(self, d):
        self.model.clear()
        self.model.update(d["model"])
        self._min_update = d["_min_update"]

    @property
    def input_dim(self):
        return len(self.model)

    def write_model(self, f, sep):
        print(sep.join(["feature"] + list(map(str, self.labels))), file=f)
        for feature, weights in self.model.items():
            print(sep.join([feature] +
                           ["%.8f" % w for w in weights.weights]), file=f)



import os
import sys
import time
from collections import OrderedDict

import numpy as np

import dynet as dy
from classifiers.classifier import Classifier
from classifiers.classifier import ClassifierProperty
from features.feature_params import MISSING_VALUE
from tupa.config import Config

TRAINERS = {
    "sgd": dy.SimpleSGDTrainer,
    "momentum": dy.MomentumSGDTrainer,
    "adagrad": dy.AdagradTrainer,
    "adadelta": dy.AdadeltaTrainer,
    "adam": dy.AdamTrainer,
}

INITIALIZERS = {
    "glorot_uniform": dy.GlorotInitializer(),
    "normal": dy.NormalInitializer(),
    "uniform": dy.UniformInitializer(1),
    "const": dy.ConstInitializer(0),
}

ACTIVATIONS = {
    "square": dy.square,
    "cube": dy.cube,
    "tanh": dy.tanh,
    "sigmoid": dy.logistic,
    "relu": dy.rectify,
}


class NeuralNetwork(Classifier):
    """
    Neural network to be used by the parser for action classification. Uses dense features.
    Keeps weights in constant-size matrices. Does not allow adding new features on-the-fly.
    Allows adding new labels on-the-fly, but requires pre-setting maximum number of labels.
    Expects features from FeatureEnumerator.
    """

    def __init__(self, *args, input_params):
        """
        Create a new untrained NN
        :param labels: a list of labels that can be updated later to add a new label
        :param input_params: dict of feature type name -> FeatureInformation
        """
        super(NeuralNetwork, self).__init__(*args)
        self.max_num_labels = Config().args.max_labels
        self._layers = Config().args.layers
        self._layer_dim = Config().args.layer_dim
        self._activation_str = Config().args.activation
        self._init_str = Config().args.init
        self._minibatch_size = Config().args.minibatch_size
        self._dropout = Config().args.dropout
        self._optimizer_str = Config().args.optimizer
        self._activation = ACTIVATIONS[self._activation_str]
        self._init = INITIALIZERS[self._init_str]
        self._optimizer = TRAINERS[self._optimizer_str]
        self._num_labels = self.num_labels
        self._params = OrderedDict()
        self._empty_values = OrderedDict()
        self._input_params = input_params
        self._indexed_num = None
        self._indexed_dim = None
        self._losses = []
        self._iteration = 0
        self._trainer = None
        self._value = None  # For caching the result of _evaluate

    @property
    def input_dim(self):
        return sum(f.num * f.dim for f in self._input_params.values())

    def resize(self):
        assert self.num_labels <= self.max_num_labels, "Exceeded maximum number of labels"

    def init_model(self):
        self.model = dy.Model()
        self._trainer = self._optimizer(self.model, )
        input_dim = self.init_input_params()
        self.init_mlp_params(input_dim)
        self.init_cg()

    def init_input_params(self):
        """
        Initialize lookup parameters and any other parameters that process the input (e.g. LSTMs)
        :return: total output dimension of inputs
        """
        input_dim = 0
        self._indexed_dim = 0
        self._indexed_num = 0
        for suffix, param in sorted(self._input_params.items()):
            if param.dim:
                if not param.numeric:  # lookup feature
                    p = self.model.add_lookup_parameters((param.size, param.dim))
                    p.set_updated(param.updated)
                    if param.init is not None:
                        p.init_from_array(param.init)
                    self._params[suffix] = p
                if param.indexed:
                    self._indexed_dim += param.dim  # add to the input dimensionality at each indexed time point
                    self._indexed_num = max(self._indexed_num, param.num)  # indices to be looked up are collected
                else:
                    input_dim += param.num * param.dim
        return input_dim + self.init_indexed_input_params()

    def init_indexed_input_params(self):
        """
        :return: total output dimension of indexed features
        """
        return self._indexed_dim * self._indexed_num

    def init_mlp_params(self, input_dim):
        for i in range(1, self._layers + 1):
            in_dim = input_dim if i == 1 else self._layer_dim
            out_dim = self._layer_dim if i < self._layers else self.max_num_labels
            self._params["W%d" % i] = self.model.add_parameters((out_dim, in_dim), init=self._init)
            self._params["b%d" % i] = self.model.add_parameters(out_dim, init=self._init)

    def init_cg(self):
        dy.renew_cg()
        for suffix, param in sorted(self._input_params.items()):
            if not param.numeric and param.dim:  # lookup feature
                self._empty_values[suffix] = self.zero_input(param.dim)

    @staticmethod
    def zero_input(dim):
        """
        Representation for missing elements
        :param dim: dimension of vector to return
        :return: zero vector (an alternative could be to learn this value, as in e.g. Kiperwasser and Goldberg 2016)
        """
        return dy.inputVector(np.zeros(dim, dtype=float))

    def generate_inputs(self, features):
        indices = []  # list, not set, in order to maintain consistent order
        for suffix, values in sorted(features.items()):
            param = self._input_params[suffix]
            if param.numeric:
                yield dy.inputVector(values)
            elif param.dim:
                if param.indexed:  # collect indices to be looked up
                    indices += values  # FeatureIndexer collapsed the features so there are no repetitions between them
                else:
                    yield dy.concatenate([self._empty_values[suffix] if x == MISSING_VALUE else self._params[suffix][x]
                                          for x in values])
        if indices:
            assert len(indices) == self._indexed_num, "Wrong number of index features: %d != %d" % (
                len(indices), self._indexed_num)
            yield self.index_input(indices)

    def index_input(self, indices):
        """
        :param indices: indices of inputs
        :return: feature values at given indices
        """
        raise Exception("Input representations not initialized, cannot evaluate indexed features")

    def evaluate_mlp(self, features, train=False):
        """
        Apply MLP and log softmax to input features
        :param features: dictionary of suffix, values for each feature type
        :param train: whether to apply dropout
        :return: expression corresponding to log softmax applied to MLP output
        """
        x = dy.concatenate(list(self.generate_inputs(features)))
        for i in range(1, self._layers + 1):
            W = dy.parameter(self._params["W%d" % i])
            b = dy.parameter(self._params["b%d" % i])
            if train and self._dropout:
                x = dy.dropout(x, self._dropout)
            x = self._activation(W * x + b)
        return dy.log_softmax(x)#, restrict=list(range(self.num_labels)))

    def evaluate(self, *args, **kwargs):
        if self.model is None:
            self.init_model()
        if self._value is None:
            self._value = self.evaluate_mlp(*args, **kwargs)
        return self._value

    def score(self, features):
        """
        Calculate score for each label
        :param features: extracted feature values, of size input_size
        :return: array with score for each label
        """
        super(NeuralNetwork, self).score(features)
        if self._iteration > 0:
            return self.evaluate(features).npvalue()[:self.num_labels]
        else:
            if Config().args.verbose >= 2:
                print("  no updates done yet, returning zero vector.")
            return np.zeros(self.num_labels)

    def update(self, features, pred, true, importance=1):
        """
        Update classifier weights according to predicted and true labels
        :param features: extracted feature values, of size input_size
        :param pred: label predicted by the classifier (non-negative integer less than num_labels)
        :param true: true label (non-negative integer less than num_labels)
        :param importance: add this many samples with the same features
        """
        super(NeuralNetwork, self).update(features, pred, true, importance)
        for _ in range(int(importance)):
            self._losses.append(dy.pick(self.evaluate(features, train=True), true))
            if Config().args.dynet_viz:
                dy.print_graphviz()
                sys.exit(0)

    def finished_step(self, train=False):
        self._value = None

    def finished_item(self, train=False):
        if len(self._losses) >= self._minibatch_size:
            self.finalize()
        elif not train:
            self.init_cg()
        self.finished_step(train)

    def finalize(self, finished_epoch=False):
        """
        Fit this model on collected samples
        :return self
        """
        super(NeuralNetwork, self).finalize()
        if self.model is None:
            self.init_model()
        if self._losses:
            loss = -dy.esum(self._losses)
            loss.forward()
            if Config().args.verbose >= 2:
                print("Total loss from %d time steps: %g" % (len(self._losses), loss.value()))
            loss.backward()
            self._trainer.update()
            self.init_cg()
            self._losses = []
            self._iteration += 1
        if finished_epoch:
            self._trainer.update_epoch()
        if Config().args.verbose:
            self._trainer.status()
        return self

    def save_model(self):
        self.finalize()
        d = {
            "input_params": self._input_params,
            "param_keys": list(self._params.keys()),
            "layers": self._layers,
            "layer_dim": self._layer_dim,
            "activation": self._activation_str,
            "init": self._init_str,
            "optimizer": self._optimizer_str,
            "iteration": self._iteration,
        }
        d.update(self.save_extra())
        model_filename = self.filename + ".model"
        started = time.time()
        try:
            os.remove(model_filename)
            print("Removed existing '%s'." % model_filename)
        except OSError:
            pass
        print("Saving model to '%s'... " % model_filename, end="", flush=True)
        try:
            self.model.save(model_filename, self._params.values())
            print("Done (%.3fs)." % (time.time() - started))
        except ValueError as e:
            print("Failed saving model: %s" % e)
        return d

    def load_model(self, d):
        self.init_model()
        self._input_params = d["input_params"]
        param_keys = d["param_keys"]
        self._layers = d["layers"]
        self._layer_dim = d["layer_dim"]
        self._activation_str = d["activation"]
        self._activation = ACTIVATIONS[self._activation_str]
        self._init_str = d["init"]
        self._init = INITIALIZERS[self._init_str]
        self._optimizer_str = d["optimizer"]
        self._optimizer = TRAINERS[self._optimizer_str]
        self._iteration = d.get("iteration", 0)
        self.load_extra(d)
        model_filename = self.filename + ".model"
        print("Loading model from '%s'... " % model_filename, end="", flush=True)
        started = time.time()
        try:
            param_values = self.model.load(model_filename)
            print("Done (%.3fs)." % (time.time() - started))
            self._params = OrderedDict(zip(param_keys, param_values))
        except KeyError as e:
            print("Failed loading model: %s" % e)

    def get_classifier_properties(self):
        return super(NeuralNetwork, self).get_classifier_properties() + \
               (ClassifierProperty.trainable_after_saving,)



import dynet as dy
from classifiers.classifier import ClassifierProperty
from features.feature_params import MISSING_VALUE
from nn.neural_network import NeuralNetwork
from tupa.config import Config, BILSTM_NN


class BiLSTM(NeuralNetwork):

    def __init__(self, *args, **kwargs):
        super(BiLSTM, self).__init__(BILSTM_NN, *args, **kwargs)
        self._lstm_layers = Config().args.lstm_layers
        self._lstm_layer_dim = Config().args.lstm_layer_dim
        self._embedding_layers = Config().args.embedding_layers
        self._embedding_layer_dim = Config().args.embedding_layer_dim
        self._input_reps = None
        self._empty_rep = None

    def init_indexed_input_params(self):
        """
        Initialize BiLSTM builder
        :return: total output dimension of BiLSTM
        """
        for i in range(1, self._embedding_layers + 1):
            in_dim = self._indexed_dim if i == 1 else self._embedding_layer_dim
            out_dim = self._embedding_layer_dim if i < self._embedding_layers else self._lstm_layer_dim
            self._params["W%de" % i] = self.model.add_parameters((out_dim, in_dim), init=self._init)
            self._params["b%de" % i] = self.model.add_parameters(out_dim, init=self._init)
        self._params["bilstm"] = dy.BiRNNBuilder(self._lstm_layers,
                                                 self._lstm_layer_dim if self._embedding_layers else self._indexed_dim,
                                                 self._lstm_layer_dim, self.model, dy.LSTMBuilder)
        return self._indexed_num * self._lstm_layer_dim

    def init_features(self, features, train=False):
        if self.model is None:
            self.init_model()
        embeddings = [[self._params[s][k] for k in ks] for s, ks in sorted(features.items())]  # time-lists of vectors
        inputs = [self.evaluate_embeddings(e, train=train) for e in zip(*embeddings)]  # join each time step to a vector
        bilstm = self._params["bilstm"]
        if train:
            bilstm.set_dropout(self._dropout)
        else:
            bilstm.disable_dropout()
        self._input_reps = bilstm.transduce(inputs)
        self._empty_rep = self.zero_input(self._lstm_layer_dim)

    def evaluate_embeddings(self, embeddings, train=False):
        """
        Apply MLP to process a single time-step of embeddings to prepare input for LSTM
        :param embeddings: list of embedding features for a single time step
        :param train: whether to apply dropout
        :return: expression corresponding to MLP output
        """
        x = dy.concatenate(list(embeddings))
        for i in range(1, self._embedding_layers + 1):
            W = dy.parameter(self._params["W%de" % i])
            b = dy.parameter(self._params["b%de" % i])
            if train and self._dropout:
                x = dy.dropout(x, self._dropout)
            x = self._activation(W * x + b)
        return x

    def index_input(self, indices):
        """
        :param indices: indices of inputs
        :return: BiLSTM outputs at given indices
        """
        return dy.concatenate([self._empty_rep if i == MISSING_VALUE else self._input_reps[i] for i in indices])

    def save_extra(self):
        return {
            "lstm_layers": self._lstm_layers,
            "lstm_layer_dim": self._lstm_layer_dim,
        }

    def load_extra(self, d):
        self._lstm_layers = d["lstm_layers"]
        self._lstm_layer_dim = d["lstm_layer_dim"]

    def get_classifier_properties(self):
        return super(BiLSTM, self).get_classifier_properties() + \
               (ClassifierProperty.require_init_features,)



from features.feature_extractor import FeatureExtractor
from features.feature_params import copy_params, NumericFeatureParameters
from tupa.model_util import load_dict, save_dict, UnknownDict
from ucca.textutil import get_word_vectors


class FeatureExtractorWrapper(FeatureExtractor):
    def __init__(self, feature_extractor, params):
        """
        :param feature_extractor: DenseFeatureExtractor to wrap
        :param params: list of FeatureParameters
        """
        super().__init__(())
        self.feature_extractor = feature_extractor
        self.params = params if isinstance(params, dict) else self.init_params(params)

    def init_params(self, param_list):
        params = {p.suffix: p for p in param_list}
        param = NumericFeatureParameters(self.feature_extractor.num_features_numeric())
        params[param.suffix] = param
        return params

    def init_features(self, state):
        return {s: self.feature_extractor.init_features(state, s)
                for s, p in self.params.items() if p.indexed and p.dim}

    def extract_features(self, state):
        return self.feature_extractor.extract_features(state)

    def finalize(self):
        return self.__class__(self.feature_extractor, copy_params(self.params, UnknownDict))

    def save(self, filename):
        save_dict(filename + self.filename_suffix(), copy_params(self.params))

    def load(self, filename):
        self.params.update(copy_params(load_dict(filename + self.filename_suffix()), UnknownDict))

    def filename_suffix(self):
        return self.feature_extractor.filename_suffix()

    @staticmethod
    def get_word_vectors(param):
        vectors, param.dim = get_word_vectors(param.dim, param.size, param.filename)
        if param.size is None:
            param.size = len(vectors)
        else:
            assert len(vectors) == param.size, "Number of loaded vectors differs from requested: %d != %d" % (
                len(vectors), param.size)
        return vectors



import numpy as np

from linear.perceptron import Perceptron
from tupa.config import DENSE_PERCEPTRON


class DensePerceptron(Perceptron):
    """
    Multi-class averaged perceptron for dense features.
    Keeps weights in a constant-size matrix. Does not allow adding new features on-the-fly.
    Allows adding new labels on-the-fly.
    Expects features from FeatureEmbedding.
    """

    def __init__(self, *args, model=None, num_features=None, epoch=0):
        """
        Create a new untrained Perceptron or copy the weights from an existing one
        :param labels: a list of labels that can be updated later to add a new label
        :param num_features: number of features that will be used for the matrix size
        :param model: if given, copy the weights (from a trained model)
        """
        super(DensePerceptron, self).__init__(DENSE_PERCEPTRON, *args, model=model, epoch=epoch)
        if not self.is_frozen:
            self._num_labels = self.num_labels
            self.input_dim = num_features
            self.model = np.zeros((self.input_dim, self.num_labels), dtype=float)
            self._totals = np.zeros((self.input_dim, self.num_labels), dtype=float)
            self._last_update = np.zeros(self.num_labels, dtype=int)

    def score(self, features):
        """
        Calculate score for each label
        :param features: extracted feature values, of size num_features
        :return: array with score for each label
        """
        super(DensePerceptron, self).score(features)
        return self.model.T.dot(features).reshape((-1,))

    def update(self, features, pred, true, importance=1):
        """
        Update classifier weights according to predicted and true labels
        :param features: extracted feature values, of size num_features
        :param pred: label predicted by the classifier (non-negative integer less than num_labels)
        :param true: true label (non-negative integer less than num_labels)
        :param importance: how much to scale the feature vector for the weight update
        """
        super(DensePerceptron, self).update(features, pred, true, importance)
        self._update(pred, -importance * self.learning_rate * features)
        self._update(true, importance * self.learning_rate * features)

    def _update(self, label, values):
        self._update_totals(label)
        self.model[:, label] += values.reshape((-1,))

    def _update_totals(self, label=None):
        self._totals[:, label] += self.model[:, label] * (self._update_index - self._last_update[label])
        self._last_update[label] = self._update_index

    def resize(self):
        self.model.resize((self.input_dim, self.num_labels), refcheck=False)
        self._totals.resize((self.input_dim, self.num_labels), refcheck=False)
        self._last_update.resize(self.num_labels, refcheck=False)

    def _finalize_model(self, average):
        model = self._totals / self._update_index if average else self.model
        return DensePerceptron(self.filename, list(self.labels), model=model, epoch=self.epoch)

    def write_model(self, f, sep):
        print(list(map(str, self.labels)), file=f)
        for row in self.model:
            print(sep.join(["%.8f" % w for w in row]), file=f)



from features.feature_extractor import FeatureExtractor
from features.feature_params import MISSING_VALUE

NON_NUMERIC_FEATURE_SUFFIXES = "wtdepxA"
FEATURE_TEMPLATES = (
    # words
    "s0ws1ws2ws3w"  # stack
    "b0wb1wb2wb3w"  # buffer
    "s0lws0rws0uws1lws1rws1uw"  # children
    "s0llws0lrws0luws0rlws0rrws0ruws0ulws0urw"
    "s0uuws1llws1lrws1luws1rlws1rrws1ruw"  # grandchildren
    "s0Uws1Uwb0Uw",  # parents
    # POS tags
    "s0ts1ts2ts3t"  # stack
    "b0tb1tb2tb3t",  # buffer
    # dependency relations
    "s0ds1ds2ds3d"  # stack
    "b0db1db2db3d",  # buffer
    # edge tags
    "s0es1es2es3e"  # stack
    "s0les0res0ues1les1res1ue"  # children
    "s0lles0lres0lues0rles0rres0rues0ules0ure"
    "s0uues1lles1lres1lues1rles1rres1rue"  # grandchildren
    "s0Ues1Ueb0Ue"  # parents
    "s0b0eb0s0e"  # specific edges
    "a0ea1e",  # past actions edge tags
    # past action labels
    "a0Aa1A",
    # separators
    "s0ps1p",
    # gap types
    "s0xs1xs2xs3x",
    # numeric
    "s0hqyPCIRs1hqys2hys3hy"
    "b0hPCIR"
    "s0s1s0b0s0",
)
EXTRA_NUMERIC_FEATURES = 2  # bias, node ratio


class DenseFeatureExtractor(FeatureExtractor):
    """
    Object to extract features from the parser state to be used in action classification
    Requires wrapping by FeatureEmbedding.
    To be used with DensePerceptron or NeuralNetwork classifier.
    """
    def __init__(self, feature_templates=FEATURE_TEMPLATES):
        super(DenseFeatureExtractor, self).__init__(feature_templates=feature_templates)
        self.numeric_features_template = None
        self.non_numeric_feature_templates = []
        for feature_template in self.feature_templates:
            if feature_template.suffix in NON_NUMERIC_FEATURE_SUFFIXES:
                self.non_numeric_feature_templates.append(feature_template)
            else:
                assert self.numeric_features_template is None, \
                    "More than one numeric feature template: %s and %s" % (
                        self.numeric_features_template, feature_template)
                self.numeric_features_template = feature_template
        self.non_numeric_by_suffix = {}
        for feature_template in self.non_numeric_feature_templates:
            for element in feature_template.elements:
                assert len(element.properties) <= 1,\
                    "Non-numeric element with %d properties: %s in feature %s" % (
                        len(element.properties), element, feature_template)
                if element.properties:
                    assert element.properties == feature_template.elements[0].properties, \
                        "Non-uniform feature template element properties: %s, %s" % (
                            element, feature_template.elements[0])
            assert feature_template.suffix not in self.non_numeric_by_suffix, \
                "More than one non-numeric feature with '%s' suffix: %s and %s" % (
                    feature_template.suffix,
                    self.non_numeric_by_suffix[feature_template.suffix],
                    feature_template)
            self.non_numeric_by_suffix[feature_template.suffix] = feature_template

    def init_features(self, state, suffix):
        return [self.get_prop(None, n, None, suffix, state) for n in state.terminals]

    def extract_features(self, state, params=None):
        """
        Calculate feature values according to current state
        :param state: current state of the parser
        :param params: dict of FeatureParameters for each suffix
        :return pair: (list of values for all numeric features,
                       list of (suffix, value) pairs for all non-numeric features)
        """
        numeric_features = [1, state.node_ratio()] + \
            self.calc_feature(self.numeric_features_template, state, default=0)
        non_numeric_features = {}
        for f in self.non_numeric_feature_templates:
            indexed = params is not None and params[f.suffix].indexed
            non_numeric_features[f.suffix] = self.calc_feature(
                f, state, default=MISSING_VALUE if indexed else "", indexed=indexed)
        # assert len(numeric_features) == self.num_features_numeric(), \
        #     "Invalid number of numeric features: %d != %d" % (
        #         len(numeric_features), self.num_features_numeric())
        # for value, element in zip(numeric_features, self.numeric_features_template.elements):
        #     assert isinstance(value, Number), \
        #         "Non-numeric value %s for numeric feature element %s" % (value, element)
        # for values, template in zip(non_numeric_features, self.non_numeric_feature_templates):
        #     for value, element in zip(values, template.elements):
        #         assert not isinstance(value, Number), \
        #             "Numeric value %s for non-numeric feature element %s" % (value, element)
        return numeric_features, non_numeric_features

    def collapse_features(self, suffixes):
        longest_suffix = max(suffixes, key=lambda s: len(self.non_numeric_by_suffix[s].elements))
        longest = self.non_numeric_by_suffix[longest_suffix]
        for suffix in suffixes:
            if suffix != longest_suffix:
                template = self.non_numeric_by_suffix.get(suffix)
                if template is not None:
                    template.elements = [e for e in template.elements if e not in longest.elements]

    def num_features_numeric(self):
        assert self.numeric_features_template is not None, \
            "Missing numeric features template"
        return sum([len(e.properties) for e in self.numeric_features_template.elements] +
                   [len([e for e in self.numeric_features_template.elements
                         if not e.properties])]) - 1 + EXTRA_NUMERIC_FEATURES

    def num_features_non_numeric(self, suffix):
        feature_template = self.non_numeric_by_suffix.get(suffix)
        assert feature_template is not None, \
            "Missing feature template for suffix '%s'" % suffix
        return sum(len(e.properties) for e in feature_template.elements)

    def features_exist(self, suffix):
        template = self.numeric_features_template if suffix == "numeric" else self.non_numeric_by_suffix.get(suffix)
        return template is not None



from tupa.config import Config
from ucca.layer1 import EdgeTags


class Constraints:
    # A unit may not have more than one outgoing edge with the same tag, if it is one of these:
    UniqueOutgoing = {
        EdgeTags.LinkRelation,
        EdgeTags.Process,
        EdgeTags.State,
    }

    # A unit may not have more than one incoming edge with the same tag, if it is one of these:
    UniqueIncoming = {
        EdgeTags.Function,
        EdgeTags.Ground,
        EdgeTags.ParallelScene,
        EdgeTags.Linker,
        EdgeTags.LinkRelation,
        EdgeTags.Connector,
        EdgeTags.Punctuation,
        EdgeTags.Terminal,
    }

    # A unit may not have more than one outgoing edge with any of these:
    MutuallyExclusiveOutgoing = {
        EdgeTags.Process,
        EdgeTags.State,
    }

    # A unit may not have any children if it has any of these incoming edge tags:
    ChildlessIncoming = {
        EdgeTags.Function,
    }

    # A childless unit may still have these outgoing edge tags:
    ChildlessOutgoing = {
        EdgeTags.Terminal,
        EdgeTags.Punctuation,
    }

    # A unit with any outgoing edge with one of these tags is a scene:
    SceneSufficientOutgoing = {
        EdgeTags.Participant,
        EdgeTags.Process,
        EdgeTags.State,
    }

    # A scene unit must have any outgoing edge with one of these tags:
    SceneNecessaryOutgoing = {
        EdgeTags.Process,
        EdgeTags.State,
    }

    # A unit with any incoming edge with one of these tags is a scene:
    SceneSufficientIncoming = {
        EdgeTags.ParallelScene,
    }

    # Outgoing edges from the root may only have these tags:
    TopLevel = {
        EdgeTags.ParallelScene,
        EdgeTags.Linker,
        EdgeTags.Function,
        EdgeTags.Ground,
        EdgeTags.Punctuation,
    }

    # A linker may only have incoming edges with these tags, and must have both:
    LinkerIncoming = {
        EdgeTags.Linker,
        EdgeTags.LinkRelation,
    }

    # Only a unit with one of these incoming tags may also have another non-remote incoming edge:
    PossibleMultipleIncoming = {
        EdgeTags.LinkArgument,
        EdgeTags.LinkRelation,
    }

    @classmethod
    def possible_multiple_incoming(cls):
        return cls.PossibleMultipleIncoming if Config().args.linkage else ()



import os
import time
from enum import Enum

from classifiers.classifier import ClassifierProperty
from tupa.action import Actions
from tupa.config import Config
from tupa.model import Model
from tupa.oracle import Oracle
from states.state import State
from ucca import diffutil, ioutil, textutil, evaluation, layer0, layer1


class ParserException(Exception):
    pass


class ParseMode(Enum):
    train = 1
    dev = 2
    test = 3


class Parser(object):

    """
    Main class to implement transition-based UCCA parser
    """
    def __init__(self, model_file=None, model_type=None, beam=1):
        self.state = None  # State object created at each parse
        self.oracle = None  # Oracle object created at each parse
        self.scores = None  # NumPy array of action scores at each action
        self.action_count = 0
        self.correct_count = 0
        self.total_actions = 0
        self.total_correct = 0
        self.model = Model(model_type, model_file, Actions().all)
        self.beam = beam  # Currently unused
        self.state_hash_history = None  # For loop checking
        # Used in verify_passage to optionally ignore a mismatch in linkage nodes:
        self.ignore_node = None if Config().args.linkage else lambda n: n.tag == layer1.NodeTags.Linkage
        self.best_score = self.dev = self.iteration = self.eval_index = None
        self.dev_scores = []
        self.trained = False

    def train(self, passages=None, dev=None, iterations=1):
        """
        Train parser on given passages
        :param passages: iterable of passages to train on
        :param dev: iterable of passages to tune on
        :param iterations: number of iterations to perform
        :return: trained model
        """
        self.trained = True
        if passages:
            if ClassifierProperty.trainable_after_saving in self.model.model.get_classifier_properties():
                try:
                    self.model.load()
                except FileNotFoundError:
                    print("not found, starting from untrained model.")
            self.best_score = 0
            self.dev = dev
            if Config().args.devscores:
                with open(Config().args.devscores, "w") as f:
                    print(",".join(["iteration"] + evaluation.Scores.field_titles(Config().args.constructions)), file=f)
            for self.iteration in range(1, iterations + 1):
                self.eval_index = 0
                print("Training iteration %d of %d: " % (self.iteration, iterations))
                list(self.parse(passages, mode=ParseMode.train))
                self.eval_and_save(self.iteration == iterations, finished_epoch=True)
                Config().random.shuffle(passages)
            print("Trained %d iterations" % iterations)
        if dev or not passages:
            self.model.load()

    def eval_and_save(self, last=False, finished_epoch=False):
        model = self.model
        self.model = self.model.finalize(finished_epoch=finished_epoch)
        if self.dev:
            print("Evaluating on dev passages")
            scores = [s for _, s in self.parse(self.dev, mode=ParseMode.dev, evaluate=True)]
            scores = evaluation.Scores.aggregate(scores)
            self.dev_scores.append(scores)
            score = scores.average_f1()
            print("Average labeled F1 score on dev: %.3f" % score)
            if Config().args.devscores:
                prefix = [self.iteration]
                if Config().args.save_every:
                    prefix.append(self.eval_index)
                with open(Config().args.devscores, "a") as f:
                    print(",".join([".".join(map(str, prefix))] + scores.fields()), file=f)
            if score >= self.best_score:
                print("Better than previous best score (%.3f)" % self.best_score)
                self.best_score = score
                self.model.save()
            else:
                print("Not better than previous best score (%.3f)" % self.best_score)
        elif last or Config().args.save_every is not None:
            self.model.save()
        if not last:
            self.model = model  # Restore non-finalized model

    def parse(self, passages, mode=ParseMode.test, evaluate=False):
        """
        Parse given passages
        :param passages: iterable of passages to parse
        :param mode: ParseMode value.
                     If train, use oracle to train on given passages.
                     Otherwise, just parse with classifier.
        :param evaluate: whether to evaluate parsed passages with respect to given ones.
                         Only possible when given passages are annotated.
        :return: generator of parsed passages (or in train mode, the original ones),
                 or, if evaluate=True, of pairs of (Passage, Scores).
        """
        assert mode in ParseMode, "Invalid parse mode: %s" % mode
        train = (mode is ParseMode.train)
        if not train and not self.trained:
            self.train()
        passage_word = "sentence" if Config().args.sentences else \
                       "paragraph" if Config().args.paragraphs else \
                       "passage"
        self.total_actions = 0
        self.total_correct = 0
        total_duration = 0
        total_tokens = 0
        passage_index = 0
        if not hasattr(passages, "__iter__"):  # Single passage given
            passages = (passages,)
        for passage_index, passage in enumerate(passages):
            l0 = passage.layer(layer0.LAYER_ID)
            num_tokens = len(l0.all)
            l1 = passage.layer(layer1.LAYER_ID)
            labeled = len(l1.all) > 1
            assert not train or labeled, "Cannot train on unannotated passage: %s" % passage.ID
            assert not evaluate or labeled, "Cannot evaluate on unannotated passage: %s" % passage.ID
            print("%s %-7s" % (passage_word, passage.ID), end=Config().line_end, flush=True)
            started = time.time()
            self.action_count = 0
            self.correct_count = 0
            textutil.annotate(passage, verbose=Config().args.verbose)  # tag POS and parse dependencies
            self.state = State(passage)
            self.state_hash_history = set()
            self.oracle = Oracle(passage) if train else None
            failed = False
            if ClassifierProperty.require_init_features in self.model.model.get_classifier_properties():
                self.model.init_features(self.state, train)
            try:
                self.parse_passage(train)  # This is where the actual parsing takes place
            except ParserException as e:
                if train:
                    raise
                Config().log("%s %s: %s" % (passage_word, passage.ID, e))
                failed = True
            predicted_passage = self.state.create_passage(assert_proper=Config().args.verify) \
                if not train or Config().args.verify else passage
            duration = time.time() - started
            total_duration += duration
            num_tokens -= len(self.state.buffer)
            total_tokens += num_tokens
            if train:  # We have an oracle to verify by
                if not failed and Config().args.verify:
                    self.verify_passage(passage, predicted_passage, train)
                if self.action_count:
                    print("%-16s" % ("%d%% (%d/%d)" %
                          (100 * self.correct_count / self.action_count,
                           self.correct_count, self.action_count)), end=Config().line_end)
            print("%0.3fs" % duration, end="")
            print("%-15s" % (" (failed)" if failed else " (%d tokens/s)" % (num_tokens / duration)), end="")
            print(Config().line_end, end="")
            if train:
                print(Config().line_end, flush=True)
            self.model.model.finished_item(train)
            self.total_correct += self.correct_count
            self.total_actions += self.action_count
            if train and Config().args.save_every and (passage_index+1) % Config().args.save_every == 0:
                self.eval_and_save()
                self.eval_index += 1
            yield (predicted_passage, evaluate_passage(predicted_passage, passage)) if evaluate else predicted_passage

        if passages:
            print("Parsed %d %ss" % (passage_index+1, passage_word))
            if self.oracle and self.total_actions:
                print("Overall %d%% correct transitions (%d/%d) on %s" %
                      (100 * self.total_correct / self.total_actions,
                       self.total_correct, self.total_actions,
                       mode.name))
            print("Total time: %.3fs (average time/%s: %.3fs, average tokens/s: %d)" % (
                total_duration, passage_word, total_duration / (passage_index+1),
                total_tokens / total_duration), flush=True)

    def parse_passage(self, train):
        """
        Internal method to parse a single passage
        :param train: use oracle to train on given passages, or just parse with classifier?
        """
        if Config().args.verbose:
            print("  initial state: %s" % self.state)
        while True:
            if Config().args.check_loops:
                self.check_loop(print_oracle=train)

            true_actions = []
            if self.oracle is not None:
                try:
                    true_actions = self.oracle.get_actions(self.state)
                except (AttributeError, AssertionError) as e:
                    if train:
                        raise ParserException("Error in oracle during training") from e

            features = self.model.feature_extractor.extract_features(self.state)
            predicted_action = self.predict_action(features, true_actions)  # sets self.scores
            action = predicted_action
            correct_action = False
            if not true_actions:
                true_actions = "?"
            elif predicted_action in true_actions:
                self.correct_count += 1
                correct_action = True
            elif train:
                action = Config().random.choice(true_actions)
            if train and not (correct_action and
                              ClassifierProperty.update_only_on_error in self.model.model.get_classifier_properties()):
                best_true_action = true_actions[0] if len(true_actions) == 1 else \
                    true_actions[self.scores[[a.id for a in true_actions]].argmax()]
                self.model.model.update(features, predicted_action.id, best_true_action.id,
                                        Config().args.swap_importance if best_true_action.is_swap else 1)
            self.action_count += 1
            self.model.model.finished_step(train)
            try:
                self.state.transition(action)
            except AssertionError as e:
                raise ParserException("Invalid transition (%s): %s" % (action, e)) from e
            if Config().args.verbose:
                if self.oracle is None:
                    print("  action: %-15s %s" % (action, self.state))
                else:
                    print("  predicted: %-15s true: %-15s taken: %-15s %s" % (
                        predicted_action, "|".join(map(str, true_actions)), action, self.state))
                for line in self.state.log:
                    print("    " + line)
            if self.state.finished or train and not correct_action and Config().args.early_update:
                return  # action is FINISH

    def check_loop(self, print_oracle):
        """
        Check if the current state has already occurred, indicating a loop
        :param print_oracle: whether to print the oracle in case of an assertion error
        """
        h = hash(self.state)
        assert h not in self.state_hash_history,\
            "\n".join(["Transition loop", self.state.str("\n")] +
                      [self.oracle.str("\n")] if print_oracle else ())
        self.state_hash_history.add(h)

    def predict_action(self, features, true_actions):
        """
        Choose action based on classifier
        :param features: extracted feature values
        :param true_actions: from the oracle, to copy orig_node if the same action is selected
        :return: valid action with maximum probability according to classifier
        """
        self.scores = self.model.model.score(features)  # Returns a NumPy array
        if Config().args.verbose >= 2:
            print("  scores: " + " ".join(("%g" % s for s in self.scores)))
        best_action = self.select_action(self.scores.argmax(), true_actions)
        if self.state.is_valid(best_action):
            return best_action
        # Usually the best action is valid, so max is enough to choose it in O(n) time
        # Otherwise, sort all the other scores to choose the best valid one in O(n lg n)
        sorted_ids = self.scores.argsort()[::-1]
        actions = (self.select_action(i, true_actions) for i in sorted_ids)
        try:
            return next(a for a in actions if self.state.is_valid(a))
        except StopIteration as e:
            raise ParserException("No valid actions available\n" +
                                  ("True actions: %s" % true_actions if true_actions
                                   else self.oracle.log if self.oracle is not None
                                   else "") +
                                  "\nReturned actions: %s" %
                                  [self.select_action(i) for i in sorted_ids] +
                                  "\nScores: %s" % self.scores
                                  ) from e

    @staticmethod
    def select_action(i, true_actions=()):
        """
        Find action with the given ID in true actions (if exists) or in all actions
        :param i: ID to lookup
        :param true_actions: preferred set of actions to look in first
        :return: Action with id=i
        """
        try:
            return next(a for a in true_actions if a.id == i)
        except StopIteration:
            return Actions().all[i]

    def verify_passage(self, passage, predicted_passage, show_diff):
        """
        Compare predicted passage to true passage and die if they differ
        :param passage: true passage
        :param predicted_passage: predicted passage to compare
        :param show_diff: if passages differ, show the difference between them?
                          Depends on predicted_passage having the original node IDs annotated
                          in the "remarks" field for each node.
        """
        assert passage.equals(predicted_passage, ignore_node=self.ignore_node),\
            "Failed to produce true passage" + \
            (diffutil.diff_passages(
                    passage, predicted_passage) if show_diff else "")


def train_test(train_passages, dev_passages, test_passages, args, model_suffix=""):
    """
    Train and test parser on given passage
    :param train_passages: passage to train on
    :param dev_passages: passages to evaluate on every iteration
    :param test_passages: passages to test on after training
    :param args: extra argument
    :param model_suffix: string to append to model filename before file extension
    :return: pair of (test scores, list of dev scores per iteration) where each one is a Scores object
    """
    test_scores = None
    model_base, model_ext = os.path.splitext(args.model or "ucca_" + args.classifier)
    p = Parser(model_file=model_base + model_suffix + model_ext, model_type=args.classifier, beam=args.beam)
    p.train(train_passages, dev=dev_passages, iterations=args.iterations)
    if test_passages:
        if args.train or args.folds:
            print("Evaluating on test passages")
        passage_scores = []
        evaluate = args.evaluate or train_passages
        for result in p.parse(test_passages, evaluate=evaluate):
            if evaluate:
                guessed_passage, score = result
                passage_scores.append(score)
            else:
                guessed_passage = result
                print()
            if guessed_passage is not None and not args.no_write:
                ioutil.write_passage(guessed_passage, output_format=args.format, binary=args.binary,
                                     outdir=args.outdir, prefix=args.prefix)
        if passage_scores and (not args.verbose or len(passage_scores) > 1):
            test_scores = evaluation.Scores.aggregate(passage_scores)
            print("\nAverage labeled F1 score on test: %.3f" % test_scores.average_f1())
            print("Aggregated scores:")
            test_scores.print()
            if Config().args.testscores:
                with open(Config().args.testscores, "a") as f:
                    print(",".join(test_scores.fields()), file=f)
    return test_scores, p.dev_scores


def evaluate_passage(guessed_passage, ref_passage):
    score = evaluation.evaluate(guessed_passage, ref_passage, constructions=Config().args.constructions,
                                verbose=Config().args.verbose and guessed_passage is not None)
    print("F1=%.3f" % score.average_f1(), flush=True)
    return score


# noinspection PyTypeChecker,PyStringFormat
def main():
    args = Config().args
    assert args.passages or args.train, "Either passages or --train is required (use -h for help)"
    assert args.model or args.train or args.folds, "Either --model or --train or --folds is required"
    assert not (args.train or args.dev) or args.folds is None, "--train and --dev are incompatible with --folds"
    assert args.train or not args.dev, "--dev is only possible together with --train"
    print("Running parser with %s" % Config())
    test_scores = None
    dev_scores = None
    if Config().args.testscores:
        with open(Config().args.testscores, "w") as f:
            print(",".join(evaluation.Scores.field_titles(Config().args.constructions)), file=f)
    if args.folds is not None:
        fold_scores = []
        all_passages = list(ioutil.read_files_and_dirs(args.passages,
                                                       Config().args.sentences, Config().args.paragraphs))
        assert len(all_passages) >= args.folds, \
            "%d folds are not possible with only %d passages" % (args.folds, len(all_passages))
        Config().random.shuffle(all_passages)
        folds = [all_passages[i::args.folds] for i in range(args.folds)]
        for i in range(args.folds):
            print("Fold %d of %d:" % (i + 1, args.folds))
            dev_passages = folds[i]
            test_passages = folds[(i + 1) % args.folds]
            train_passages = [passage for fold in folds
                              if fold is not dev_passages and fold is not test_passages
                              for passage in fold]
            s, _ = train_test(train_passages, dev_passages, test_passages, args, "_%d" % i)
            if s is not None:
                fold_scores.append(s)
        if fold_scores:
            test_scores = evaluation.Scores.aggregate(fold_scores)
            print("Average labeled test F1 score for each fold: " + ", ".join(
                "%.3f" % s.average_f1() for s in fold_scores))
            print("Aggregated scores across folds:\n")
            test_scores.print()
    else:  # Simple train/dev/test by given arguments
        train_passages, dev_passages, test_passages = [ioutil.read_files_and_dirs(arg,
                                                       Config().args.sentences, Config().args.paragraphs) for arg in
                                                       (args.train, args.dev, args.passages)]
        test_scores, dev_scores = train_test(train_passages, dev_passages, test_passages, args)
    return test_scores, dev_scores


if __name__ == "__main__":
    main()
    Config().close()



from features.feature_params import FeatureParameters
from tupa.action import Actions
from tupa.config import Config, SPARSE_PERCEPTRON, DENSE_PERCEPTRON, MLP_NN, BILSTM_NN


class Model(object):
    def __init__(self, model_type, filename, labels=None, feature_extractor=None, model=None):
        if model_type is None:
            model_type = SPARSE_PERCEPTRON
        self.model_type = model_type
        self.filename = filename
        if feature_extractor is not None and model is not None:
            self.feature_extractor = feature_extractor
            self.model = model
            return

        if model_type == SPARSE_PERCEPTRON:
            from features.sparse_features import SparseFeatureExtractor
            from linear.sparse_perceptron import SparsePerceptron
            self.feature_extractor = SparseFeatureExtractor()
            self.model = SparsePerceptron(filename, labels)
        elif model_type == DENSE_PERCEPTRON:
            from features.embedding import FeatureEmbedding
            from linear.dense_perceptron import DensePerceptron
            self.feature_extractor = self.dense_features_wrapper(FeatureEmbedding)
            self.model = DensePerceptron(filename, labels, num_features=self.feature_extractor.num_features())
        elif model_type == MLP_NN:
            from features.enumerator import FeatureEnumerator
            from nn.feedforward import MLP
            self.feature_extractor = self.dense_features_wrapper(FeatureEnumerator)
            self.model = MLP(filename, labels, input_params=self.feature_extractor.params)
        elif model_type == BILSTM_NN:
            from features.enumerator import FeatureEnumerator
            from features.indexer import FeatureIndexer
            from nn.bilstm import BiLSTM
            self.feature_extractor = FeatureIndexer(self.dense_features_wrapper(FeatureEnumerator))
            self.model = BiLSTM(filename, labels, input_params=self.feature_extractor.params)
        else:
            raise ValueError("Invalid model type: '%s'" % model_type)

    @staticmethod
    def dense_features_wrapper(wrapper):
        from features.dense_features import DenseFeatureExtractor
        params = [
            FeatureParameters("W", Config().args.word_dim_external, Config().args.max_words_external,
                              Config().args.word_dropout_external, Config().args.update_word_vectors, copy_from="w",
                              filename=Config().args.word_vectors),
            FeatureParameters("w", Config().args.word_dim, Config().args.max_words, Config().args.word_dropout),
            FeatureParameters("t", Config().args.tag_dim, Config().args.max_tags),
            FeatureParameters("d", Config().args.dep_dim, Config().args.max_deps),
            FeatureParameters("e", Config().args.label_dim, Config().args.max_edge_labels),
            FeatureParameters("p", Config().args.punct_dim, Config().args.max_puncts),
            FeatureParameters("x", Config().args.gap_dim, Config().args.max_gaps),
            FeatureParameters("A", Config().args.action_dim, Config().args.max_actions),
        ]
        return wrapper(DenseFeatureExtractor(), params)

    def init_features(self, state, train):
        self.model.init_features(self.feature_extractor.init_features(state), train)

    def finalize(self, finished_epoch):
        return Model(model_type=self.model_type,
                     filename=self.filename,
                     feature_extractor=self.feature_extractor.finalize(),
                     model=self.model.finalize(finished_epoch))

    def save(self):
        if self.filename is not None:
            try:
                self.feature_extractor.save(self.filename)
                self.model.save()
            except Exception as e:
                raise IOError("Failed saving model to '%s'" % self.filename) from e

    def load(self):
        if self.filename is not None:
            try:
                self.feature_extractor.load(self.filename)
                self.model.load()
                Actions().all = self.model.labels
            except FileNotFoundError:
                raise
            except Exception as e:
                raise IOError("Failed loading model from '%s'" % self.filename) from e



from collections import defaultdict

import numpy as np

from features.feature_extractor_wrapper import FeatureExtractorWrapper
from features.feature_params import MISSING_VALUE
from features.feature_params import NumericFeatureParameters
from tupa.config import Config
from tupa.model_util import UnknownDict


class FeatureEmbedding(FeatureExtractorWrapper):
    """
    Wrapper for DenseFeatureExtractor to replace non-numeric features with embeddings
    and return an array of numbers rather than (name, value) pairs.
    To be used with DensePerceptron classifier.
    """
    def __init__(self, feature_extractor, params):
        super(FeatureEmbedding, self).__init__(feature_extractor, params)

    def init_data(self, param):
        if param.data is not None or isinstance(param, NumericFeatureParameters):
            return
        param.num = self.feature_extractor.num_features_non_numeric(param.effective_suffix)
        if param.dim:
            if param.external:
                vectors = self.get_word_vectors(param)
                param.data = UnknownDict(vectors, np.zeros(param.dim))
            else:
                param.data = defaultdict(lambda d=param.dim: Config().random.normal(size=d))
                _ = param.data[UnknownDict.UNKNOWN]  # Initialize unknown value
        param.empty = np.zeros(param.dim, dtype=float)

    def extract_features(self, state):
        """
        Calculate feature values according to current state
        :param state: current state of the parser
        :return array of concatenated numeric and embedding features
        """
        numeric_features, non_numeric_features = self.feature_extractor.extract_features(state)
        features = [np.array(numeric_features, dtype=float)]
        for suffix, param in sorted(self.params.items()):
            if param.dim:
                values = non_numeric_features.get(param.effective_suffix)
                if values is not None:
                    self.init_data(param)
                    features += [param.empty if v == MISSING_VALUE else param.data[v] for v in values]
        # assert sum(map(len, features)) == self.num_features(),\
        #     "Invalid total number of features: %d != %d " % (
        #         sum(map(len, features)), self.num_features())
        return np.hstack(features).reshape((-1, 1))

    def num_features(self):
        ret = 0
        for param in self.params.values():
            if param.dim:
                self.init_data(param)
                ret += param.dim * param.num
        return ret

    def filename_suffix(self):
        return "_embedding"



#!/usr/bin/env python

from distutils.core import setup
import os

setup(name='TUPA',
      version='1.0',
      description='Transition-based UCCA Parser',
      author='Daniel Hershcovich',
      author_email='danielh@cs.huji.ac.il',
      url='http://www.cs.huji.ac.il/~oabend/ucca.html',
      packages=['ucca', 'tupa', 'classifiers', 'nn', 'linear', 'features', 'states'],
      package_dir={
          'ucca': os.path.join('ucca', 'ucca'),
          'tupa': 'tupa',
          'classifiers': os.path.join('tupa', 'classifiers'),
          'linear': os.path.join('tupa', 'classifiers', 'linear'),
          'nn': os.path.join('tupa', 'classifiers', 'nn'),
          'features': os.path.join('tupa', 'features'),
          'states': os.path.join('tupa', 'states'),
          },
      )



from nn.neural_network import NeuralNetwork
from tupa.config import MLP_NN


class MLP(NeuralNetwork):

    def __init__(self, *args, **kwargs):
        super(MLP, self).__init__(MLP_NN, *args, **kwargs)



import csv
import os
from collections import OrderedDict

import numpy as np

from tupa import parse, config
from tupa.config import Config
from ucca.evaluation import Scores

MODELS_DIR = "models"


class Params(object):
    def __init__(self, params):
        self.params = params
        self.params["model"] = "%s/ucca-%s-%d" % (MODELS_DIR, self.params["classifier"], self.params["seed"])
        self.test_scores = None
        self.dev_scores = ()

    def run(self):
        assert Config().args.train and (Config().args.passages or Config().args.dev) or \
               Config().args.passages and Config().args.folds, "insufficient parameters given to parser"
        print("Running with %s" % self)
        Config().update(self.params)
        self.test_scores, self.dev_scores = parse.main()

    def score(self):
        return self.test_scores.average_f1() if self.test_scores is not None else \
            self.dev_scores[-1].average_f1() if self.dev_scores else -float("inf")

    def __str__(self):
        ret = ", ".join("%s: %s" % (name, value) for name, value in self.params.items())
        if self.test_scores is not None:
            ret += ", average labeled f1: %.3f" % self.score()
        return ret

    def get_fields(self):
        return [str(p) for p in self.params.values()] + [str(self.score())] + self.test_scores.fields()

    def get_field_titles(self):
        return [p for p in self.params.keys()] + ["average_labeled_f1"] + Scores.field_titles()

    def write_scores(self, writer):
        for i, scores in enumerate(self.dev_scores):
            writer.writerow([str((i + 1) if n == "iterations" else p) for n, p in self.params.items()] +
                            [str(scores.average_f1())] + scores.fields())
        if self.test_scores is not None and (not self.dev_scores or
                                             self.test_scores.fields() != self.dev_scores[-1].fields()):
            writer.writerow([str(p) for p in self.params.values()] +
                            [str(self.test_scores.average_f1())] + self.test_scores.fields())


def main():
    if not os.path.exists(MODELS_DIR):
        os.makedirs(MODELS_DIR)
    Config().args.no_write = True
    out_file = os.environ.get("PARAMS_FILE", "params.csv")
    word_vectors_files = [os.environ[f] for f in os.environ if f.startswith("WORD_VECTORS")]
    num = int(os.environ.get("PARAMS_NUM", 30))
    np.random.seed()
    domains = (
        ("seed",                    2147483647),  # max value for int
        ("classifier",              (config.MLP_NN, config.BILSTM_NN)),
        ("update_word_vectors",     [True, False]),
        ("word_vectors",            [None] + word_vectors_files),
        ("word_dim_external",       (0, 300)),
        ("word_dim",                (0, 50, 100, 200, 300)),
        ("tag_dim",                 (5, 10, 20)),
        ("dep_dim",                 (5, 10, 20)),
        ("label_dim",               (5, 10, 20)),
        ("punct_dim",               (1, 2, 3)),
        ("gap_dim",                 (1, 2, 3)),
        ("action_dim",              (3, 5, 10)),
        ("layer_dim",               (50, 100, 200, 300, 500, 1000)),
        ("layers",                  [1] + 5 * [2]),
        ("lstm_layer_dim",          (50, 100, 200, 300, 500, 1000)),
        ("lstm_layers",             [1] + 5 * [2]),
        ("embedding_layer_dim",     (50, 100, 200, 300, 500, 1000)),
        ("embedding_layers",        5 * [1] + [2]),
        ("activation",              config.ACTIVATIONS),
        ("init",                    5 * [config.INITIALIZATIONS[0]] + list(config.INITIALIZATIONS)),
        ("batch_size",              (10, 30, 50, 100, 200, 500)),
        ("minibatch_size",          (50, 100, 200, 300, 500, 1000)),
        ("optimizer",               5 * [config.OPTIMIZERS[0]] + list(config.OPTIMIZERS)),
        ("swap_importance",         (1, 2)),
        ("iterations",              range(1, 51)),
        ("word_dropout",            (0, .1, .2, .25, .3)),
        ("word_dropout_external",   (0, .1, .2, .25, .3)),
        ("dynet_weight_decay",      (1e-7, 1e-6, 1e-5, 1e-4)),
        ("dropout",                 (0, .1, .2, .3, .4, .5)),
    )
    params = [Params(OrderedDict(p))
              for p in zip(*[[(n, v.item() if hasattr(v, "item") else v)
                              for v in np.random.choice(vs, num)]
                             for n, vs in domains])]
    print("All parameter combinations to try:")
    print("\n".join(map(str, params)))
    print("Saving results to '%s'" % out_file)
    with open(out_file, "w") as f:
        csv.writer(f).writerow(params[0].get_field_titles())
    for param in params:
        param.run()
        with open(out_file, "a") as f:
            param.write_scores(csv.writer(f))
        best = max(params, key=Params.score)
        print("Best parameters: %s" % best)


if __name__ == "__main__":
    main()
